/*
    rtl/com.c

*/
#define COBJMACROS

#include <rtl/com.h>

#include <olectl.h>
#include <strsafe.h>

#pragma comment (lib, "strsafe.lib")

/* -------------------------------------------------------------------------- */

#define MAX(a, b)                           (((a) > (b)) ? (a) : (b))
#define MIN(a, b)                           (((a) < (b)) ? (a) : (b))

#define OBJECT_PTR(type, object, offset)    (type)(((LPBYTE)(object)) + (offset))
#define OBJECT_FROM_FIELD(ptr, type, field) ((type*)(((LPBYTE)(ptr)) - COM_FIELD_OFFSET (type, field)))

#define POINTER_SIZE                        ((DWORD)sizeof(LPVOID))
#define NEGATIVE_ONE                        ((DWORD)-1)

#define COM_FLAGS_VALID     0x80000000  /*     */

/* -------------------------------------------------------------------------- */

typedef struct _IObjectIntf
{
    LPCVOID lpVtbl;     /*     */
    const IID* iid;     /*   */
    COM_OBJECT* object; /* ,    */
    LPVOID data;        /*  ,    */
}IObjectIntf;

typedef struct _TYPEINFONODE
{
    COM_TREE_NODE node;

    GUID id;
    ITypeInfo* typeInfo;
}TYPEINFONODE;

typedef struct _TYPELIBNODE
{
    COM_TREE_NODE node;

    GUID id;
    WORD majorVersion;
    WORD minorVersion;
    LCID lcid;
    ITypeLib* typeLib;
    COM_TREE typeInfoTree; /* TYPEINFONODE* */
}TYPELIBNODE;

enum INTERFACE_DATA_TYPE
{
    IDT_TYPEINFO,
};

typedef struct _INTFDATANODE
{
    COM_TREE_NODE node;
    int type;
}INTFDATANODE;

typedef struct _INTFTYPEINFONODE
{
    INTFDATANODE dataNode;
    LCID lcid;
    ITypeInfo* typeInfo;
}INTFTYPEINFONODE;

typedef struct _GUID_NODE   GUID_NODE;
typedef struct _IGuidObject IGuidObject;

struct _GUID_NODE
{
    COM_TREE_NODE node;

    IGuidObject* object;
};

struct _IGuidObject
{
    IGuid object;
    volatile DWORD refCount;

    GUID guid;
    GUID_NODE* node;
};

enum SERVER_LOCK_ID
{
    EVENTS_LOCK,
    OBJECT_LOCK,
    TYPE_LOCK,
    GUID_LOCK,

    NUMBER_OF_SERVER_LOCK_IDS
};

typedef struct _COM_SERVER
{
    COM_STATS stats;

    COM_EVENTS events;
    LPVOID eventsParam;

    CRITICAL_SECTION locks[NUMBER_OF_SERVER_LOCK_IDS];

    TYPELIBNODE* defaultTypeLib;
    COM_TREE typeLibTree; /* TYPELIBNODE* */

    COM_LIST runningClassList;

    COM_TREE guidTree; /* GUIDNODE* */

    IGlobalInterfaceTable* GIT;
}COM_SERVER;

static COM_SERVER g_Server = {0};

static HANDLE g_Heap = NULL;

static void COM_EnterLock (int index);
static void COM_LeaveLock (int index);

#define COM_EnterGuid() COM_EnterLock (GUID_LOCK)
#define COM_LeaveGuid() COM_LeaveLock (GUID_LOCK)

/* -------------------------------------------------------------------------- */
/*                 ,                        */
/* -------------------------------------------------------------------------- */

#define _New(type)              (type*)_Alloc (sizeof(type))
#define _Delete(ptr, type)      _Free (ptr, sizeof(type))

static LPVOID _Alloc (DWORD size)
{
    LPVOID data;

    if (!g_Heap)
        g_Heap = GetProcessHeap ();

    data = HeapAlloc (g_Heap, HEAP_ZERO_MEMORY, size);
    if (data)
        g_Server.stats.memoryUsed += size;

    return data;
}

/* -------------------------------------------------------------------------- */

static void _Clear (LPVOID data, DWORD size)
{
    memset (data, 0, size);
}

/* -------------------------------------------------------------------------- */

static LPVOID _Realloc (LPVOID ptr, DWORD size, DWORD newSize)
{
    if (!ptr)
    {
        ptr = _Alloc (newSize);
    }
    else if (size != newSize)
    {
        if (!g_Heap)
            g_Heap = GetProcessHeap();

        ptr = HeapReAlloc (g_Heap, HEAP_ZERO_MEMORY, ptr, newSize);
        if (ptr)
        {
            if (newSize > size)
            {
                g_Server.stats.memoryUsed += (newSize - size);
            }
            else
            {
                size -= newSize;

                if (g_Server.stats.memoryUsed < size)
                    g_Server.stats.memoryUsed = 0;
                else
                    g_Server.stats.memoryUsed -= size;
            }
        }
    }

    return ptr;
}

/* -------------------------------------------------------------------------- */

static void _Free (LPVOID ptr, DWORD size)
{
    if (ptr)
    {
        if (!g_Heap)
            g_Heap = GetProcessHeap ();

        HeapFree (g_Heap, 0, ptr);

        if (g_Server.stats.memoryUsed < size)
            g_Server.stats.memoryUsed = 0;
        else
            g_Server.stats.memoryUsed -= size;
    }
}

/* -------------------------------------------------------------------------- */

static void _Copy (LPVOID dest, LPCVOID src, DWORD size)
{
    memcpy (dest, src, size);
}

/* -------------------------------------------------------------------------- */

static int _Compare (LPCVOID data1, LPCVOID data2, DWORD size)
{
    return memcmp (data1, data2, size);
}

/* -------------------------------------------------------------------------- */

static DWORD _Length (LPCSTR string)
{
    return (string ? lstrlenA (string) : 0);
}

/* -------------------------------------------------------------------------- */

static void _Assign (LPSTR buffer, DWORD cchMax, LPCSTR string)
{
    if (buffer)
    {
        if (string)
            StringCchCopyA (buffer, cchMax, string);
        else
            buffer[0] = 0;
    }
}

/* -------------------------------------------------------------------------- */

static void _Append (LPSTR buffer, DWORD cchMax, LPCSTR string)
{
    if (buffer && string)
        StringCchCatA (buffer, cchMax, string);
}

/* -------------------------------------------------------------------------- */

static void _VFormatString (LPSTR buffer, DWORD cchMax, LPCSTR format, va_list args)
{
    if (buffer && format)
        StringCchVPrintfA (buffer, cchMax, format, args);
}

/* -------------------------------------------------------------------------- */

static void _FormatString (LPSTR buffer, DWORD cchMax, LPCSTR format, ...)
{
    va_list args;

    if (buffer && format)
    {
        va_start (args, format);
        _VFormatString (buffer, cchMax, format, args);
        va_end (args);
    }
}

/* -------------------------------------------------------------------------- */

static DWORD _AtomicInc (DWORD volatile* value)
{
    return (DWORD)InterlockedIncrement ((LONG*)value);
}

/* -------------------------------------------------------------------------- */

static DWORD _AtomicDec (DWORD volatile* value)
{
    return (DWORD)InterlockedDecrement ((LONG*)value);
}

/* -------------------------------------------------------------------------- */
/*                                                           */
/* -------------------------------------------------------------------------- */

static void FIRE_OnErrorMessage (LPCSTR message)
{
    COM_EnterEvents();
    
    if (g_Server.events.OnErrorMessage)
        g_Server.events.OnErrorMessage (message, g_Server.eventsParam);

    COM_LeaveEvents();
}

/* -------------------------------------------------------------------------- */

static void FIRE_OnIntegerOverflow (LPCSTR context/*=NULL*/, LPCSTR name)
{
    COM_EnterEvents();

   if (g_Server.events.OnIntegerOverflow)
       g_Server.events.OnIntegerOverflow (context, name, g_Server.eventsParam);

   COM_LeaveEvents();
}

/* -------------------------------------------------------------------------- */

static void FIRE_OnNegativeCount (LPCSTR context/*=NULL*/, LPCSTR name)
{
    COM_EnterEvents();

    if (g_Server.events.OnNegativeCount)
        g_Server.events.OnNegativeCount (context, name, g_Server.eventsParam);

    COM_LeaveEvents();
}

/* -------------------------------------------------------------------------- */

static void FIRE_OnLastObjectReleased (void)
{
    COM_EnterEvents();

    if (g_Server.events.OnLastObjectReleased)
        g_Server.events.OnLastObjectReleased (g_Server.eventsParam);

    COM_LeaveEvents();
}

/* -------------------------------------------------------------------------- */

static void FIRE_OnMemoryLeak (void)
{
    COM_EnterEvents();

    if (g_Server.events.OnMemoryLeak)
        g_Server.events.OnMemoryLeak (g_Server.eventsParam);

    COM_LeaveEvents();
}

/* -------------------------------------------------------------------------- */

static void _StringFromGUID (LPCGUID guid, LPSTR buffer)
{
    if (buffer)
    {
        if (guid)
        {
            _FormatString (buffer, 256, "%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X",
                guid->Data1,
                guid->Data2,
                guid->Data3,
                guid->Data4[0],
                guid->Data4[1],
                guid->Data4[2],
                guid->Data4[3],
                guid->Data4[4],
                guid->Data4[5],
                guid->Data4[6],
                guid->Data4[7]);
        }
        else
            _Assign (buffer, 256, "<NULL>");
    }
}

/* -------------------------------------------------------------------------- */

static LPCSTR _GetClassName (const COM_CLASS* _class, LPSTR className/*=NULL*/)
{
    static char buffer[256];

    if (!className)
        className = &buffer[0];

    if (!_class)
        _Assign (className, 256, "<NULL>");
    else if (_class->name)
        _Assign (className, 256, _class->name);
    else if (!_class->id)
        _Assign (className, 256, "<NULL>");
    else
        _StringFromGUID (_class->id, className);

    return className;
}

/* -------------------------------------------------------------------------- */

static LPCSTR _GetEntryInterfaceName (const COM_INTERFACE* entry, LPSTR intfName/*=NULL*/)
{
    static char buffer[256];

    if (!intfName)
        intfName = &buffer[0];

    if (!entry)
        _Assign (intfName, 256, "<NULL>");
    else if (entry->intfName)
        _Assign (intfName, 256, entry->intfName);
    else if (!entry->iid)
        _Assign (intfName, 256, "<NULL>");
    else
        _StringFromGUID (entry->iid, intfName);

    return intfName;
}

/* -------------------------------------------------------------------------- */

static LPCSTR _GetEntryFieldName (const COM_INTERFACE* entry, LPSTR fieldName/*=NULL*/)
{
    static char buffer[256];

    if (!fieldName)
        fieldName = &buffer[0];

    if (!entry)
        _Assign (fieldName, 256, "<NULL>");
    else if (entry->fieldName)
        _Assign (fieldName, 256, entry->fieldName);
    else
        _FormatString (fieldName, 256, "[+0x%08x]", entry->fieldOffset);

    return fieldName;
}

/* -------------------------------------------------------------------------- */

static LPCSTR _GetTypeLibName (REFGUID libid, DWORD version, LCID lcid, LPSTR libName/*=NULL*/)
{
    char buffer[256];
    LPSTR ptr;
    WORD major, minor;

    if (!libName)
        libName = &buffer[0];

    if (libid)
    {
        _StringFromGUID (libid, libName);
        ptr = libName + _Length (libName);

        major = COM_MAJOR_VERSION (version);
        minor = COM_MINOR_VERSION (minor);
        
        if (major != COM_ANY_VERSION)
        {
            _FormatString (ptr, 256, ", ver %u.", major);
            ptr += _Length (ptr);

            if (minor != COM_ANY_VERSION)
                _FormatString (ptr, 256, "%u", minor);
            else
                _Append (ptr, 256, "x");

            ptr += _Length (ptr);
        }

        if (lcid != 0)
        {
            _FormatString (ptr, 256, ", locale 0x%X", lcid);
            ptr += _Length (ptr);
        }
    }
    else
        _Assign (libName, 256, "<NULL>");

    return libName;
}

/* -------------------------------------------------------------------------- */

static void _ErrMsg (LPCSTR format, va_list args)
{
    char buffer[512];
    
    if (format)
    {
        _VFormatString (buffer, sizeof(buffer), format, args);
        FIRE_OnErrorMessage (buffer);
    }
}

/* -------------------------------------------------------------------------- */

static void ErrMsg (LPCSTR format, ...)
{
    va_list args;

    if (format)
    {
        va_start (args, format);
        _ErrMsg (format, args);
        va_end (args);
    }
}

/* -------------------------------------------------------------------------- */

static void ClassErrMsg (const COM_CLASS* _class, LPCSTR format, ...)
{
    va_list args;
    char newFormat[256];

    if (format)
    {
        va_start (args, format);

        if (_class)
        {
            _GetClassName (_class, newFormat);
            _Append (newFormat, sizeof(newFormat), ": ");
            _Append (newFormat, sizeof(newFormat), format);
            _ErrMsg (newFormat, args);
        }
        else
            _ErrMsg (format, args);

        va_end (args);
    }
}

/* -------------------------------------------------------------------------- */

/*
        1.

       ,     
    UINT32_MAX    COM_EVENTS::OnIntegerOverflow().

      .
*/
static DWORD IncrementCounter (LPCSTR context/*=NULL*/, LPCSTR name, DWORD volatile* value)
{
    DWORD result;

    result = _AtomicInc (value);
    if (result == 0)
    {
        result--;
        (*value)--;

        FIRE_OnIntegerOverflow (context, name);
    }

    return result;
}

/* -------------------------------------------------------------------------- */

/*
        1.

      .
*/
static DWORD DecrementCounter (DWORD volatile* value)
{
    return _AtomicDec (value);
}

/* -------------------------------------------------------------------------- */

static void DecrementObjectCount (void)
{
    DWORD count;

    count = DecrementCounter (&g_Server.stats.objectCount);

    if (count == 0)
    {
        FIRE_OnLastObjectReleased ();
    }
    else if (count == NEGATIVE_ONE)
    {
        g_Server.stats.objectCount++;

        FIRE_OnNegativeCount ("COM", "objectCount");
    }
}

/* -------------------------------------------------------------------------- */
/*                                                             */
/* -------------------------------------------------------------------------- */

static void FreeTypeInfoNode (COM_TREE* tree, COM_TREE_NODE* node)
{
    TYPEINFONODE* typeInfoNode;

    typeInfoNode = (TYPEINFONODE*)node;
    if (typeInfoNode)
    {
        if (typeInfoNode->typeInfo)
            ITypeInfo_Release (typeInfoNode->typeInfo);

        _Delete (typeInfoNode, TYPEINFONODE);
    }
}

/* -------------------------------------------------------------------------- */

static int CompareTypeInfoNodes (COM_TREE* tree, LPCVOID key, COM_TREE_NODE* node)
{
    LPGUID guid;
    TYPEINFONODE* typeInfoNode;

    guid = (LPGUID)key;
    typeInfoNode = (TYPEINFONODE*)node;

    if (guid && typeInfoNode)
        return COM_CompareGUID (guid, &typeInfoNode->id);

    return 0;
}

/* -------------------------------------------------------------------------- */

static void FreeTypeLibNode (COM_TREE* tree, COM_TREE_NODE* node)
{
    TYPELIBNODE* typeLibNode;

    typeLibNode = (TYPELIBNODE*)node;
    if (typeLibNode)
    {
        COM_ClearTree (&typeLibNode->typeInfoTree);

        if (typeLibNode->typeLib)
            ITypeLib_Release (typeLibNode->typeLib);

        _Delete (typeLibNode, TYPELIBNODE);
    }
}

/* -------------------------------------------------------------------------- */

static int CompareUINT32 (DWORD value1, DWORD value2)
{
    if (value1 < value2)
        return -1;

    if (value1 > value2)
        return 1;

    return 0;
}

/* -------------------------------------------------------------------------- */

static int CompareTypeLibNodes (COM_TREE* tree, LPCVOID key, COM_TREE_NODE* node)
{
    TYPELIBNODE *typeLibNode1, *typeLibNode2;
    int result;

    typeLibNode1 = (TYPELIBNODE*)key;
    typeLibNode2 = (TYPELIBNODE*)node;

    if (typeLibNode1 && typeLibNode2)
    {
        result = COM_CompareGUID (&typeLibNode1->id, &typeLibNode2->id);
        if (result == 0)
        {
            result = CompareUINT32 (typeLibNode1->majorVersion, typeLibNode2->majorVersion);
            if (result == 0)
            {
                result = CompareUINT32 (typeLibNode1->minorVersion, typeLibNode2->minorVersion);
                if (result == 0)
                    result = CompareUINT32 (typeLibNode1->lcid, typeLibNode2->lcid);
            }
        }

        return result;
    }

    return 0;
}

/* -------------------------------------------------------------------------- */

/*
        TYPELIBNODE      
     .
*/
static TYPELIBNODE* CreateTypeLibNode (ITypeLib* typeLib)
{
    TYPELIBNODE* typeLibNode;
    TLIBATTR* libAttr;

    typeLibNode = NULL;

    if (typeLib)
    {
        libAttr = NULL;
        ITypeLib_GetLibAttr (typeLib, &libAttr);

        if (libAttr)
        {
            typeLibNode = _New (TYPELIBNODE);
            if (typeLibNode)
            {
                COM_CopyGUID (&typeLibNode->id, &libAttr->guid);
                typeLibNode->lcid = libAttr->lcid;
                typeLibNode->majorVersion = libAttr->wMajorVerNum;
                typeLibNode->minorVersion = libAttr->wMinorVerNum;

                typeLibNode->typeInfoTree.Free = FreeTypeInfoNode;
                typeLibNode->typeInfoTree.Compare = CompareTypeInfoNodes;

                ITypeLib_AddRef (typeLib);
                typeLibNode->typeLib = typeLib;
            }

            ITypeLib_ReleaseTLibAttr (typeLib, libAttr);
        }
    }

    return typeLibNode;
}

/* -------------------------------------------------------------------------- */

/*
    ,    ,   typeLibNode, 
     version.

     TRUE, :
        -        ;
        -    version  COM_ANY_VERSION;
        -    version     
                version  COM_ANY_VERSION.
*/
static BOOL TestTypeLibVersion (TYPELIBNODE* typeLibNode, DWORD version)
{
    WORD major, minor;

    major = COM_MAJOR_VERSION (version);
    minor = COM_MINOR_VERSION (version);

    return ((major == COM_ANY_VERSION) || ((major == typeLibNode->majorVersion) && (minor == COM_ANY_VERSION || minor == typeLibNode->minorVersion)));
}

/* -------------------------------------------------------------------------- */

/*
      TYPELIBNODE      
    ,       .

      libId  NULL,    , 
        .

         COM_ANY_VERSION,   
            
    .

         COM_ANY_VERSION,   
              .

     lcid       .
*/
static TYPELIBNODE* FindTypeLibNode (LPCGUID libId/*=NULL*/, DWORD version, LCID lcid)
{
    TYPELIBNODE tmp, *typeLibNode, *parent;

    if (!libId)
    {
        typeLibNode = g_Server.defaultTypeLib;

        if (!typeLibNode || !typeLibNode->typeLib)
        {
            ErrMsg ("FindTypeLibNode(): Default type library is not set");

            return NULL;
        }
    }
    else
    {
        COM_CopyGUID (&tmp.id, libId);
        tmp.majorVersion = COM_MAJOR_VERSION (version);
        tmp.minorVersion = COM_MINOR_VERSION (version);
        tmp.lcid = lcid;

        typeLibNode = (TYPELIBNODE*)COM_FindNode (&g_Server.typeLibTree, &tmp, (COM_TREE_NODE**)&parent);
        if (!typeLibNode && parent && parent->node.left)
            typeLibNode = parent;
    }

    if (typeLibNode && typeLibNode->typeLib && TestTypeLibVersion (typeLibNode, version) && typeLibNode->lcid == lcid)
        return typeLibNode;

    return NULL;
}

/* -------------------------------------------------------------------------- */

/*
      TYPEINFONODE     typeId 
    ,   typeLibNode.

          ,    
    ITypeLib::GetTypeInfoOfGuid().
*/
static TYPEINFONODE* GetTypeInfoNode (TYPELIBNODE* typeLibNode, REFGUID typeId)
{
    TYPEINFONODE *typeInfoNode;
    char libName[256];
    char typeName[256];
    HRESULT hr;

    if (typeLibNode && typeLibNode->typeInfoTree.root && typeId)
    {
        typeInfoNode = (TYPEINFONODE*)COM_FindNode (&typeLibNode->typeInfoTree, typeId, NULL);
        if (typeInfoNode)
            return typeInfoNode;

        typeInfoNode = _New (TYPEINFONODE);
        if (typeInfoNode)
        {
            COM_CopyGUID (&typeInfoNode->id, typeId);

            hr = ITypeLib_GetTypeInfoOfGuid (typeLibNode->typeLib, typeId, &typeInfoNode->typeInfo);
            if (SUCCEEDED (hr))
            {
                if (COM_AddNode (&typeLibNode->typeInfoTree, typeId, &typeInfoNode->node))
                    return typeInfoNode;
            }
            else if (hr == TYPE_E_ELEMENTNOTFOUND)
            {
                _StringFromGUID (typeId, typeName);
                _GetTypeLibName (&typeLibNode->id, 
                    COM_MAKE_VERSION (typeLibNode->majorVersion, typeLibNode->minorVersion), 
                    typeLibNode->lcid, 
                    libName);

                ErrMsg ("GetTypeInfoNode(): Type %s not found in library (%s)", typeName, libName);
            }

            _Delete (typeInfoNode, TYPEINFONODE);
        }
    }

    return NULL;
}

/* -------------------------------------------------------------------------- */
/*                                              */
/* -------------------------------------------------------------------------- */

static DWORD AbsDiff (DWORD value1, DWORD value2)
{
    return ((value1 > value2) ? (value1 - value2) : (value2 - value1));
}

/* -------------------------------------------------------------------------- */

static BOOL IsDescendant (const COM_CLASS* _class, const COM_CLASS* parent)
{
    if (_class && parent)
        for (_class = _class->parent; _class; _class = _class->parent)
            if (_class == parent)
                return TRUE;

    return FALSE;
}

/* -------------------------------------------------------------------------- */

static BOOL IsValidClass (const COM_CLASS* _class)
{
    COM_RUNTIME_CLASS* runtimeClass;
    const COM_INTERFACE* table;
    const COM_INTERFACE* entry;
    const COM_INTERFACE* prev;
    DWORD minOffset, maxOffset;
    char className[256];
    char fieldName[256];

    if (!_class)
    {
        ErrMsg ("Invalid parameter: class is NULL");

        return FALSE;
    }

    runtimeClass = (COM_RUNTIME_CLASS*)_class->runtimeClass;
    if (runtimeClass)
    {
        if (runtimeClass->flags & COM_FLAGS_VALID)
        {
            /*       */

            return TRUE;
        }

        runtimeClass->_class = _class;
        runtimeClass->listItem.prev = NULL;
        runtimeClass->listItem.next = NULL;
        runtimeClass->objectList.head = NULL;
        runtimeClass->objectList.tail = NULL;
        runtimeClass->objectList.Free = NULL;
    }

    if (!_class->id)
    {
        ClassErrMsg (_class, "Class GUID is NULL");

        return FALSE;
    }

    if (_class->name && _Length (_class->name) >= 256)
    {
        ClassErrMsg (_class, "Class name is too long");

        return FALSE;
    }

    _GetClassName (_class, className);

    if (_class->parent)
    {
        if (_class == _class->parent || IsDescendant (_class->parent, _class))
        {
            ClassErrMsg (_class, "Cyclic inheritance detected");

            return FALSE;
        }

        if (!IsValidClass (_class->parent))
        {
            ClassErrMsg (_class, "Invalid parent class");

            return FALSE;
        }

        if (_class->instanceSize < _class->parent->instanceSize)
        {
            ClassErrMsg (_class, "Invalid instance size");

            return FALSE;
        }

        minOffset = _class->parent->instanceSize;
        maxOffset = _class->instanceSize - POINTER_SIZE;
    }
    else
    {
        if (_class->instanceSize < sizeof(COM_OBJECT))
        {
            ClassErrMsg (_class, "Invalid instance size");

            return FALSE;
        }

        minOffset = (DWORD)sizeof(COM_OBJECT);
        maxOffset = _class->instanceSize - POINTER_SIZE;
    }

    entry = table = &_class->interfaces[0];
    while (entry->type != CIT_END)
    {
        _GetEntryFieldName (entry, fieldName);

        if (entry->type == CIT_GENERIC || 
            entry->type == CIT_AGGREGATE || 
            entry->type == CIT_AGGREGATE2)
        {
            if (entry->fieldSize != POINTER_SIZE)
            {
                ClassErrMsg (_class, "Invalid interface entry field %s::%s %s", className, fieldName, "(wrong size)");

                return FALSE;
            }

            if (entry->fieldOffset < minOffset || entry->fieldOffset > maxOffset)
            {
                ClassErrMsg (_class, "Invalid interface entry field %s::%s %s", className, fieldName, "(wrong offset)");

                return FALSE;
            }

            for (prev = table; prev != entry; prev++)
            {
                if (prev->fieldOffset == entry->fieldOffset)
                {
                    if (prev->type != entry->type)
                    {
                        ClassErrMsg (_class, "Invalid interface entry field %s::%s %s", className, fieldName, "(type mismatch)");

                        return FALSE;
                    }
                }
                else if (AbsDiff (prev->fieldOffset, entry->fieldOffset) < POINTER_SIZE)
                {
                    _GetEntryFieldName (entry, fieldName);
                    ClassErrMsg (_class, "Invalid interface entry field %s::%s %s", className, fieldName, "(wrong offset)");

                    return FALSE;
                }
            }
        }

        switch (entry->type)
        {
        case CIT_GENERIC:

            if (!entry->iid)
            {
                ClassErrMsg (_class, "Invalid interface entry field %s::%s %s", className, fieldName, "(iid is NULL)");

                return FALSE;
            }

            if (!entry->lpVtbl)
            {
                ClassErrMsg (_class, "Invalid interface entry field %s::%s %s", className, fieldName, "(vtable is NULL)");

                return FALSE;
            }

            break;

        case CIT_AGGREGATE:

            if (entry->innerCLSID && COM_IsEqualGUID (entry->innerCLSID, _class->id))
            {
                ClassErrMsg (_class, "Class aggregates itself");

                return FALSE;
            }

            break;

        case CIT_AGGREGATE2:

            if (entry->innerClass && (entry->innerClass == _class || IsDescendant (entry->innerClass, _class)))
            {
                ClassErrMsg (_class, "Class aggregates itself");

                return FALSE;
            }

            break;

        case CIT_NOINTERFACE:

            if (!entry->iid)
            {
                ClassErrMsg (_class, "Invalid no-interface entry (iid is NULL)");

                return FALSE;
            }

            break;

        case CIT_CALLBACK:

            if (!entry->QI)
            {
                ClassErrMsg (_class, "Invalid callback interface entry (proc is NULL)");

                return FALSE;
            }

            break;

        default:
            ClassErrMsg (_class, "Interface entry of unknown type %u (ignored)", entry->type);
            break;

        }

        entry++;
    }

    if (runtimeClass)
        runtimeClass->flags |= COM_FLAGS_VALID;

    return TRUE;
}

/* -------------------------------------------------------------------------- */

static DWORD CalculateIntfTableSize (const COM_CLASS* _class)
{
    COM_RUNTIME_CLASS* runtimeClass;
    const COM_INTERFACE* table;
    const COM_INTERFACE* entry;
    const COM_INTERFACE* prev;
    DWORD size;

    runtimeClass = (COM_RUNTIME_CLASS*)_class->runtimeClass;

    if (runtimeClass && runtimeClass->intfTableSize != NEGATIVE_ONE)
        return runtimeClass->intfTableSize;

    size = 0;

    entry = table = &_class->interfaces[0];
    while (entry->type != CIT_END)
    {
        if (entry->type > CIT_END && entry->type < CIT_LAST)
        {
            for (prev = table; prev != entry; prev++)
                if (prev->fieldOffset == entry->fieldOffset)
                    break;

            if (prev == entry)
            {
                if (entry->type == CIT_GENERIC)
                    size += sizeof(IObjectIntf);
            }
        }

        entry++;
    }

    if (runtimeClass)
        runtimeClass->intfTableSize = size;

    return size;
}

/* -------------------------------------------------------------------------- */

static DWORD CalculateTotalObjectSize (const COM_CLASS* _class)
{
    COM_RUNTIME_CLASS* runtimeClass;
    DWORD size;

    runtimeClass = (COM_RUNTIME_CLASS*)_class->runtimeClass;

    if (runtimeClass && runtimeClass->objectSize != 0)
        return runtimeClass->objectSize;

    size = _class->instanceSize;

    while (_class)
    {
        size += CalculateIntfTableSize (_class);
        _class = _class->parent;
    }

    if (runtimeClass)
        runtimeClass->objectSize = size;

    return size;
}

/* -------------------------------------------------------------------------- */

static void FreeIntfDataNode (COM_TREE* tree, COM_TREE_NODE* node)
{
    INTFDATANODE* dataNode;
    INTFTYPEINFONODE* typeInfoNode;
    DWORD nodeSize;

    dataNode = (INTFDATANODE*)node;
    if (dataNode)
    {
        nodeSize = 0;

        switch (dataNode->type)
        {
        case IDT_TYPEINFO:

            nodeSize = sizeof(INTFTYPEINFONODE);
            typeInfoNode = (INTFTYPEINFONODE*)dataNode;

            if (typeInfoNode->typeInfo)
                ITypeInfo_Release (typeInfoNode->typeInfo);

            break;
        }

        _Free (dataNode, nodeSize);
    }
}

/* -------------------------------------------------------------------------- */

static int CompareIntfDataNodes (COM_TREE* tree, LPCVOID key, COM_TREE_NODE* node)
{
    INTFDATANODE *dataNode1, *dataNode2;
    INTFTYPEINFONODE *typeInfoNode1, *typeInfoNode2;

    dataNode1 = (INTFDATANODE*)key;
    dataNode2 = (INTFDATANODE*)node;
    if (dataNode1 && dataNode2)
    {
        if (dataNode1->type < dataNode2->type)
            return -1;

        if (dataNode1->type > dataNode2->type)
            return 1;

        switch (dataNode1->type)
        {
        case IDT_TYPEINFO:

            typeInfoNode1 = (INTFTYPEINFONODE*)dataNode1;
            typeInfoNode2 = (INTFTYPEINFONODE*)dataNode2;

            if (typeInfoNode1->lcid < typeInfoNode2->lcid)
                return -1;

            if (typeInfoNode1->lcid > typeInfoNode2->lcid)
                return 1;

            break;
        }
    }

    return 0;
}

/* -------------------------------------------------------------------------- */

static HRESULT AddInterfaceData (IObjectIntf* intf, INTFDATANODE* node)
{
    if (!intf || !node)
        return E_INVALIDARG;

    if (!intf->data)
    {
        intf->data = _New (COM_TREE);
        if (!intf->data)
            return E_OUTOFMEMORY;

        ((COM_TREE*)intf->data)->Free = FreeIntfDataNode;
        ((COM_TREE*)intf->data)->Compare = CompareIntfDataNodes;
    }

    if (!COM_AddNode ((COM_TREE*)intf->data, node, (COM_TREE_NODE*)node))
    {
        if (!(((COM_TREE*)intf->data)->root))
        {
            _Delete (intf->data, COM_TREE);
            intf->data = NULL;
        }

        return E_ALREADYEXISTS;
    }

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static BOOL AddInterfaceTypeInfo (IObjectIntf* intf, LCID lcid, ITypeInfo* typeInfo)
{
    INTFTYPEINFONODE* node;
    HRESULT hr;

    if (intf && typeInfo)
    {
        node = _New (INTFTYPEINFONODE);
        if (node)
        {
            ITypeInfo_AddRef (typeInfo);

            node->dataNode.type = IDT_TYPEINFO;
            node->lcid = lcid;
            node->typeInfo = typeInfo;

            hr = AddInterfaceData (intf, &node->dataNode);
            if (hr == S_OK)
                return TRUE;

            FreeIntfDataNode (NULL, (COM_TREE_NODE*)node);

            if (hr == E_ALREADYEXISTS)
                return TRUE;
        }
    }

    return FALSE;
}

/* -------------------------------------------------------------------------- */

static void ReleaseObjectInterface (COM_OBJECT* object, IObjectIntf** ptr)
{
    IObjectIntf* intf;

    intf = *ptr;
    if (intf)
    {
        intf->lpVtbl = NULL;
        intf->object = NULL;

        if (intf->data)
        {
            COM_ClearTree ((COM_TREE*)intf->data);
            _Delete (intf->data, COM_TREE);
            intf->data = NULL;
        }

        *ptr = NULL;
    }
}

/* -------------------------------------------------------------------------- */

static void ReleaseAggregatedInterface (COM_OBJECT* object, IUnknown** ptr)
{
    IUnknown* intf;

    intf = *ptr;
    if (intf)
    {
        IUnknown_Release (intf);

        *ptr = NULL;
    }
}

/* -------------------------------------------------------------------------- */

static void ReleaseObjectInterfaces (COM_OBJECT* object)
{
    const COM_CLASS* _class;
    const COM_INTERFACE* entry;

    _class = object->_class;
    while (_class)
    {
        entry = &_class->interfaces[0];
        while (entry->type != CIT_END)
        {
            switch (entry->type)
            {
            case CIT_GENERIC:
                ReleaseObjectInterface (object, OBJECT_PTR (IObjectIntf**, object, entry->fieldOffset));
                break;

            case CIT_AGGREGATE:
            case CIT_AGGREGATE2:
                ReleaseAggregatedInterface (object, OBJECT_PTR (IUnknown**, object, entry->fieldOffset));
                break;

            case CIT_NOINTERFACE:
            case CIT_CALLBACK:
                break;
            }

            entry++;
        }

        _class = _class->parent;
    }
}

/* -------------------------------------------------------------------------- */

static HRESULT InitGenericInterface (COM_OBJECT* object, const COM_INTERFACE* entry, IObjectIntf** ptr, LPVOID* intfTablePtr)
{
    IObjectIntf* intf;
    char className[256];
    char intfName[256];
    char fieldName[256];
    TYPELIBNODE* tli;
    TYPEINFONODE* tii;
    HRESULT hr;

    hr = S_OK;

    if (*ptr == NULL)
    {
        _GetClassName (object->_class, className);
        _GetEntryInterfaceName (entry, intfName);
        _GetEntryFieldName (entry, fieldName);

        intf = (IObjectIntf*)(*intfTablePtr);
        *intfTablePtr = ((LPBYTE)(*intfTablePtr)) + sizeof(IObjectIntf);

        intf->lpVtbl = entry->lpVtbl;
        intf->iid = entry->iid;
        intf->object = object;
        intf->data = NULL;
        
        if (entry->libID != NULL || entry->libVersion != 0 || entry->libLCID != 0)
        {
            COM_EnterType();

            hr = TYPE_E_ELEMENTNOTFOUND;

            tli = FindTypeLibNode (entry->libID, entry->libVersion, entry->libLCID);
            if (tli)
            {
                tii = GetTypeInfoNode (tli, entry->iid);
                if (tii && tii->typeInfo)
                    hr = AddInterfaceTypeInfo (intf, entry->libLCID, tii->typeInfo) ? S_OK : E_OUTOFMEMORY;
                else
                    ClassErrMsg (object->_class, "Interface %s, field %s::%s: interface not found in type library", intfName, className, fieldName);
            }
            else
                ClassErrMsg (object->_class, "Interface %s, field %s::%s: type library not found", intfName, className, fieldName);

            COM_LeaveType();
        }

        *ptr = intf;
    }

    return hr;
}

/* -------------------------------------------------------------------------- */

static HRESULT InitAggregateInterface (COM_OBJECT* object, const COM_INTERFACE* entry, IUnknown** ptr, LPVOID* intfTablePtr)
{
    IUnknown* intf;
    HRESULT hr;

    if (*ptr == NULL)
    {
        if (entry->innerClass)
        {
            intf = NULL;

            if (entry->type == CIT_AGGREGATE)
            {
                hr = CoCreateInstance (entry->innerCLSID, object->unknOuter, CLSCTX_INPROC_SERVER, &IID_IUnknown, &intf);
            }
            else if (entry->type == CIT_AGGREGATE2)
            {
                hr = COM_CreateInstance (entry->innerClass, object->unknOuter, &IID_IUnknown, &intf);
            }
            else
            {
                ClassErrMsg (object->_class, "InitAggregateInterface(): invalid entry type");

                return E_FAIL;
            }

            if (FAILED (hr) || !intf)
            {
                if (SUCCEEDED(hr))
                    hr = E_FAIL;

                ClassErrMsg (object->_class, "Failed to create aggregated object: 0x%08x", hr);

                return hr;
            }

            *ptr = intf;
        }
    }

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static HRESULT InitObjectInterfaces (COM_OBJECT* object)
{
    const COM_CLASS* _class;
    const COM_INTERFACE* entry;
    LPVOID intfTablePtr;
    HRESULT result;

    _class = object->_class;
    intfTablePtr = OBJECT_PTR (LPVOID, object, _class->instanceSize);

    while (_class)
    {
        entry = &_class->interfaces[0];
        while (entry->type != CIT_END)
        {
            switch (entry->type)
            {
            case CIT_GENERIC:
                result = InitGenericInterface (object, entry, OBJECT_PTR (IObjectIntf**, object, entry->fieldOffset), &intfTablePtr);
                break;

            case CIT_AGGREGATE:
            case CIT_AGGREGATE2:
                result = InitAggregateInterface (object, entry, OBJECT_PTR (IUnknown**, object, entry->fieldOffset), &intfTablePtr);
                break;

            case CIT_NOINTERFACE:
                result = S_OK;
                break;

            case CIT_CALLBACK:
                result = S_OK;
                break;

            default:
                result = E_INVALIDTYPE;
                break;
            }

            if (FAILED (result))
            {
                ReleaseObjectInterfaces (object);

                return result;
            }

            entry++;
        }

        _class = _class->parent;
    }

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static void _FreeObject (COM_OBJECT* object, const COM_CLASS* _class/*=NULL*/)
{
    if (object)
    {
        if (!_class)
            _class = object->_class;

        while (_class)
        {
            if (_class->FreeInstance)
                _class->FreeInstance (object);

            _class = _class->parent;
        }
    }
}

/* -------------------------------------------------------------------------- */

static void _DestroyObject (COM_OBJECT* object)
{
    COM_RUNTIME_CLASS* runtimeClass;
    char className[256];
    DWORD instanceSize;

    if (object)
    {
        instanceSize = CalculateTotalObjectSize (object->_class);

        _FreeObject (object, NULL);
        ReleaseObjectInterfaces (object);

        runtimeClass = (COM_RUNTIME_CLASS*)object->_class->runtimeClass;

        if (runtimeClass)
        {
            if (runtimeClass->flags & COM_FLAGS_SINGLETON)
                runtimeClass->singletonInstance = NULL;

            COM_RemoveItem (&runtimeClass->objectList, &object->listItem);

            if (DecrementCounter (&runtimeClass->instanceCount) == NEGATIVE_ONE)
            {
                runtimeClass->instanceCount = 0;
                _GetClassName (runtimeClass->_class, className);
                FIRE_OnNegativeCount (className, "instanceCount");
            }

            if (runtimeClass->instanceCount == 0)
            {
                COM_RemoveItem (&g_Server.runningClassList, &runtimeClass->listItem);

                if (runtimeClass->lock)
                {
                    DeleteCriticalSection (runtimeClass->lock);
                    _Delete (runtimeClass->lock, CRITICAL_SECTION);
                    runtimeClass->lock = NULL;
                }
            }
        }

        _Free (object, instanceSize);

        DecrementObjectCount();
    }
}

/* -------------------------------------------------------------------------- */

static HRESULT _InitObject (COM_OBJECT* object, const COM_CLASS* _class/*=NULL*/)
{
    const COM_INTERFACE* entry;
    COM_RUNTIME_CLASS* runtimeClass;
    IUnknown** ptr;
    HRESULT hr;

    if (!_class)
        _class = object->_class;

    if (_class->parent)
        hr = _InitObject (object, _class->parent);
    else
        hr = InitObjectInterfaces (object);

    if (hr == S_OK)
    {
        runtimeClass = (COM_RUNTIME_CLASS*)object->_class->runtimeClass;
        
        if (runtimeClass && !runtimeClass->lock)
        {
            runtimeClass->lock = _New (CRITICAL_SECTION);
            if (runtimeClass->lock)
                InitializeCriticalSection (runtimeClass->lock);
        }

        if (_class->InitInstance)
        {
            hr = _class->InitInstance (object);
            if (hr == S_OK)
            {
                entry = &object->_class->interfaces[0];
                while (entry->type != CIT_END)
                {
                    if (entry->type == CIT_AGGREGATE || entry->type == CIT_AGGREGATE2)
                    {
                        ptr = OBJECT_PTR (IUnknown**, object, entry->fieldOffset);

                        if (*ptr == NULL)
                        {
                            ClassErrMsg (object->_class, "Aggregated interface not initialized");
                            hr = E_FAIL;

                            break;
                        }
                    }

                    entry++;
                }
            }

            if (hr != S_OK)
            {
                if (_class->parent)
                    _FreeObject (object, _class->parent);
                else
                    ReleaseObjectInterfaces (object);
            }
        }

        if (hr != S_OK && runtimeClass)
        {
            if (runtimeClass->instanceCount == 0 && runtimeClass->lock)
            {
                DeleteCriticalSection (runtimeClass->lock);
                _Delete (runtimeClass->lock, CRITICAL_SECTION);
                runtimeClass->lock = NULL;
            }
        }
    }

    return hr;
}

/* -------------------------------------------------------------------------- */
/*                       IUnknown                         */
/* -------------------------------------------------------------------------- */

/*   IUnknown */
static STDMETHODIMP   CUnknown_NDQueryInterface   (IUnknown* This, REFIID riid, LPVOID* ppvObject);
static ULONG WINAPI     CUnknown_NDAddRef           (IUnknown* This);
static ULONG WINAPI     CUnknown_NDRelease          (IUnknown* This);

COM_DECLARE_VTABLE (CUnknown, IUnknown)
{
    CUnknown_NDQueryInterface,
    CUnknown_NDAddRef,
    CUnknown_NDRelease,
};

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknown_NDQueryInterface (IUnknown* This, REFIID riid, LPVOID* ppvObject)
{
    COM_OBJECT* object;
    const COM_CLASS* _class;
    const COM_INTERFACE* entry;

    IUnknown** intfPtr;

    HRESULT hr;

    object = (COM_OBJECT*)This;

    if (ppvObject)
        *ppvObject = NULL;

    if (!object || !riid || !ppvObject)
        return E_INVALIDARG;

    if (COM_IsEqualGUID (riid, &IID_IUnknown))
    {
        IUnknown_AddRef (This);
        *ppvObject = This;

        return S_OK;
    }

    _class = object->_class;
    while (_class)
    {
        entry = &_class->interfaces[0];
        while (entry->type != CIT_END)
        {
            intfPtr = OBJECT_PTR (IUnknown**, object, entry->fieldOffset);
            if (*intfPtr)
            {
                switch (entry->type)
                {
                case CIT_GENERIC:
                    
                    if (entry->iid && COM_IsEqualGUID (riid, entry->iid))
                    {
                        IUnknown_AddRef (*intfPtr);
                        *ppvObject = *intfPtr;

                        return S_OK;
                    }

                    break;

                case CIT_AGGREGATE:
                case CIT_AGGREGATE2:

                    if (!entry->iid || COM_IsEqualGUID (entry->iid, riid))
                    {
                        hr = IUnknown_QueryInterface (*intfPtr, riid, ppvObject);
                        if (SUCCEEDED (hr))
                            return S_OK;

                        if (entry->iid)
                            return hr;
                        
                        /* BLIND AGGREGATION FAILED.      */
                    }

                    break;

                case CIT_NOINTERFACE:
                    return E_NOINTERFACE;

                case CIT_CALLBACK:

                    hr = entry->QI (object, entry->dw, riid, ppvObject);
                    if (hr != S_FALSE)
                        return hr;

                    break;
                }
            }

            entry++;
        }

        _class = _class->parent;
    }

    return E_NOINTERFACE;
}

/* -------------------------------------------------------------------------- */

static ULONG WINAPI CUnknown_NDAddRef (IUnknown* This)
{
    COM_OBJECT* object;

    object = (COM_OBJECT*)This;

    if (!object)
        return 0;

    return (ULONG)IncrementCounter (object->_class->name, "refCount", &object->refCount);
}

/* -------------------------------------------------------------------------- */

static ULONG WINAPI CUnknown_NDRelease (IUnknown* This)
{
    COM_OBJECT* object;
    const COM_CLASS* _class;
    DWORD refCount;

    object = (COM_OBJECT*)This;

    if (!object)
        return 0;

    _class = object->_class;

    refCount = DecrementCounter (&object->refCount);
    if (refCount == 0)
    {
        COM_EnterObject();

        if (object->refCount != 0)
        {
            /*
                     -  
                COM_CreateObject()
            */
            refCount = object->refCount;
        }
        else
        {
            _DestroyObject (object);
        }

        COM_LeaveObject();
    }
    else if (refCount == NEGATIVE_ONE)
    {
        /*       */
        FIRE_OnNegativeCount (_class->name, "refCount");
    }

    return refCount;
}

/* -------------------------------------------------------------------------- */

HRESULT WINAPI CUnknown_QueryInterface (IUnknown* This, REFIID riid, LPVOID* ppvObject)
{
    IObjectIntf* intf;

    intf = (IObjectIntf*)This;

    if (ppvObject)
        *ppvObject = NULL;
    else
        return E_POINTER;

    if (!intf || !intf->object || !intf->iid || !riid || !ppvObject)
        return E_INVALIDARG;

    if (COM_IsEqualGUID (riid, &IID_IUnknown) || COM_IsEqualGUID (riid, intf->iid))
    {
        IUnknown_AddRef (This);
        *ppvObject = This;

        return S_OK;
    }

    return IUnknown_QueryInterface (intf->object->unknOuter, riid, ppvObject);
}

/* -------------------------------------------------------------------------- */

ULONG WINAPI CUnknown_AddRef (IUnknown* This)
{
    COM_OBJECT* object;

    object = COM_GetObject (COM_OBJECT, This);

    if (!object)
        return 0;

    return IUnknown_AddRef (object->unknOuter);
}

/* -------------------------------------------------------------------------- */

ULONG WINAPI CUnknown_Release (IUnknown* This)
{
    COM_OBJECT* object;

    object = COM_GetObject (COM_OBJECT, This);

    if (!object)
        return 0;

    return IUnknown_Release (object->unknOuter);
}

/* -------------------------------------------------------------------------- */
/*                       IDispatch                        */
/* -------------------------------------------------------------------------- */

HRESULT WINAPI CDispatch_GetTypeInfoCount (IDispatch* This, UINT* pctinfo)
{
    IObjectIntf* intf;
    COM_TREE_NODE* node;

    if (pctinfo)
        *pctinfo = 0;

    intf = (IObjectIntf*)This;
    if (!intf || !intf->object || !pctinfo)
        return E_INVALIDARG;

    COM_EnterType();

    if (intf->data)
    {
        node = COM_FirstNode ((COM_TREE*)intf->data);

        while (node && ((INTFDATANODE*)node)->type != IDT_TYPEINFO)
            node = COM_NextNode ((COM_TREE*)intf->data, node);

        if (node && ((INTFTYPEINFONODE*)node)->typeInfo)
            *pctinfo = 1;
    }

    COM_LeaveType();

    return S_OK;
}

/* -------------------------------------------------------------------------- */

HRESULT WINAPI CDispatch_GetTypeInfo (IDispatch* This, UINT iTInfo, LCID lcid, ITypeInfo** ppTInfo)
{
    if (ppTInfo)
        *ppTInfo = NULL;

    if (!This || !ppTInfo)
        return E_INVALIDARG;

    if (iTInfo != 0)
        return DISP_E_BADINDEX;

    return COM_GetInterfaceTypeInfo (This, lcid, ppTInfo);
}

/* -------------------------------------------------------------------------- */

HRESULT WINAPI CDispatch_GetIDsOfNames (IDispatch* This, REFIID riid, LPOLESTR* rgszNames, UINT cNames, LCID lcid, DISPID* rgDispId)
{
    ITypeInfo* typeInfo;
    HRESULT hr;

    if (!This || !rgszNames || !rgDispId)
        return E_INVALIDARG;

    hr = COM_GetInterfaceTypeInfo (This, lcid, &typeInfo);
    if (SUCCEEDED (hr))
    {
        hr = ITypeInfo_GetIDsOfNames (typeInfo, rgszNames, cNames, rgDispId);
        ITypeInfo_Release (typeInfo);
    }

    return hr;
}

/* -------------------------------------------------------------------------- */

HRESULT WINAPI CDispatch_Invoke (IDispatch* This, DISPID dispIdMember, REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS* pDispParams, VARIANT* pVarResult, EXCEPINFO* pExcepInfo, UINT* puArgErr)
{
    ITypeInfo* typeInfo;
    HRESULT hr;

    if (!This)
        return E_INVALIDARG;

    hr = COM_GetInterfaceTypeInfo (This, lcid, &typeInfo);
    if (SUCCEEDED (hr))
    {
        hr = ITypeInfo_Invoke (typeInfo, This, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
        ITypeInfo_Release (typeInfo);
    }

    return hr;
}

/* -------------------------------------------------------------------------- */
/*                               CEnumerator                             */
/* -------------------------------------------------------------------------- */
/*
      :

    interface IEnum<T> : public IUnknown
    {
        STDMETHOD (Next     ) (ULONG count, T[] buffer, ULONG* pFetched);
        STDMETHOD (Skip     ) (ULONG count);
        STDMETHOD (Reset    ) ();
        STDMETHOD (Clone    ) (IEnum<T>** ppEnum);
    };

        -, 
     IEnumerable.
*/

/* {2E01FC69-0718-4057-BCEA-D46011074705} */
static const GUID CLSID_CEnumerator = {0x2e01fc69, 0x718, 0x4057, {0xbc, 0xea, 0xd4, 0x60, 0x11, 0x7, 0x47, 0x5}};

typedef struct _IEnumeratorVtbl IEnumeratorVtbl;
typedef struct _IEnumerator     IEnumerator;
typedef struct _CEnumerator     CEnumerator;

struct _IEnumeratorVtbl
{
    /* IUnknown */
    STDMETHOD  (QueryInterface  ) (IEnumerator* This, REFIID iid, LPVOID* ppvObject);
    STDMETHOD_ (ULONG, AddRef   ) (IEnumerator* This);
    STDMETHOD_ (ULONG, Release  ) (IEnumerator* This);

    /* IEnum<T> */
    STDMETHOD  (Next            ) (IEnumerator* This, ULONG count, LPVOID buffer, ULONG* pFetched);
    STDMETHOD  (Skip            ) (IEnumerator* This, ULONG count);
    STDMETHOD  (Reset           ) (IEnumerator* This);
    STDMETHOD  (Clone           ) (IEnumerator* This, IEnumerator** ppEnum);
};

struct _IEnumerator
{
    const IEnumeratorVtbl* lpVtbl;
};

#ifdef COBJMACROS
#   define IEnumerator_QueryInterface(This, iid, ppvObject)     ((This)->lpVtbl->QueryInterface (This, iid, ppvObject))
#   define IEnumerator_AddRef(This)                             ((This)->lpVtbl->AddRef (This))
#   define IEnumerator_Release(This)                            ((This)->lpVtbl->Release (This))
#   define IEnumerator_Next(This, count, buffer, pFetched)      ((This)->lpVtbl->Next (This, count, buffer, pFetched))
#   define IEnumerator_Skip(This, count)                        ((This)->lpVtbl->Skip (This, count))
#   define IEnumerator_Reset(This)                              ((This)->lpVtbl->Reset (This))
#   define IEnumerator_Clone(This, ppEnum)                      ((This)->lpVtbl->Clone (This, ppEnum))
#endif

static STDMETHODIMP CEnumerator_Next    (IEnumerator* This, ULONG count, LPVOID buffer, ULONG* pFetched);
static STDMETHODIMP CEnumerator_Skip    (IEnumerator* This, ULONG count);
static STDMETHODIMP CEnumerator_Reset   (IEnumerator* This);
static STDMETHODIMP CEnumerator_Clone   (IEnumerator* This, IEnumerator** ppEnum);

COM_DECLARE_VTABLE(CEnumerator, IEnumerator)
{
    /* IUnknown */
    COM_IUNKNOWN_METHODS (IEnumerator),

    /* IEnumerator */
    CEnumerator_Next,
    CEnumerator_Skip,
    CEnumerator_Reset,
    CEnumerator_Clone,
};

struct _CEnumerator
{
    COM_OBJECT object;

    /* implemented interfaces */
    IEnumerator* intf;

    /* aggregated interfaces */

    /* private data */
    LPCRITICAL_SECTION lock;

    IGuid* iid;
    IEnumerable* enumerable;
    LPVOID iterator;
};

static HRESULT  COM_CTOR_NAME (CEnumerator) (COM_OBJECT* object);
static void     COM_DTOR_NAME (CEnumerator) (COM_OBJECT* object);

static HRESULT  CEnumerator_QueryInterface (COM_OBJECT* object, DWORD dw, REFIID iid, LPVOID* ppvObject);

BEGIN_INTERFACE_MAP(CEnumerator)
    COM_INTERFACE_ENTRY_CALLBACK (CEnumerator_QueryInterface, 0),
END_INTERFACE_MAP();

COM_DECLARE_PRIVATE_CLASS (NULL, CEnumerator, 0);

/* -------------------------------------------------------------------------- */

static HRESULT COM_CTOR_NAME (CEnumerator) (COM_OBJECT* object)
{
    CEnumerator* en;

    en = (CEnumerator*)object;

    en->lock = COM_GetClassLock (object);
    if (!en->lock)
    {
        en->lock = _New (CRITICAL_SECTION);
        if (!en->lock)
            return E_OUTOFMEMORY;

        InitializeCriticalSection (en->lock);
    }

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static void COM_DTOR_NAME (CEnumerator) (COM_OBJECT* object)
{
    CEnumerator* en;

    en = (CEnumerator*)object;

    if (en->lock && !COM_IsClassLock (object, en->lock))
    {
        DeleteCriticalSection (en->lock);
        _Delete (en->lock, CRITICAL_SECTION);
        en->lock = NULL;
    }

    if (en->iid)
    {
        IGuid_Release (en->iid);
        en->iid = NULL;
    }

    if (en->enumerable)
    {
        IEnumerable_FreeIterator (en->enumerable, en->iterator);
        IEnumerable_Release (en->enumerable);
        en->enumerable = NULL;
    }

    if (en->intf)
    {
        COM_DestroyObjectInterface (en->intf);
        en->intf = NULL;
    }
}

/* -------------------------------------------------------------------------- */

static HRESULT CEnumerator_QueryInterface (COM_OBJECT* object, DWORD dw, REFIID iid, LPVOID* ppvObject)
{
    CEnumerator* This;
    HRESULT hr;

    This = (CEnumerator*)object;

    *ppvObject = NULL;

    if (IGuid_Equal (This->iid, iid))
    {
        EnterCriticalSection (This->lock);

        if (!This->intf)
        {
            hr = COM_CreateObjectInterface (object, iid, &COM_VTABLE_NAME (CEnumerator, IEnumerator), &This->intf);
            if (hr != S_OK)
            {
                LeaveCriticalSection (This->lock);

                return hr;
            }
        }

        LeaveCriticalSection (This->lock);

        IEnumerator_AddRef (This->intf);
        *ppvObject = This->intf;

        return S_OK;
    }

    return S_FALSE;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CEnumerator_Next (IEnumerator* This, ULONG count, LPVOID buffer, ULONG* pFetched)
{
    CEnumerator* en;
    HRESULT hr;

    if (pFetched)
        *pFetched = 0;

    if (count == 0)
        return S_OK;

    en = COM_GetObject (CEnumerator, This);
    if (!en || (count && !buffer))
        return E_INVALIDARG;

    EnterCriticalSection (en->lock);

    if (en->enumerable)
        hr = IEnumerable_Next (en->enumerable, &en->iterator, count, buffer, pFetched);
    else
        hr = E_UNEXPECTED;

    LeaveCriticalSection (en->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CEnumerator_Skip (IEnumerator* This, ULONG count)
{
    CEnumerator* en;
    HRESULT hr;

    if (count == 0)
        return S_OK;

    en = COM_GetObject (CEnumerator, This);
    if (!en)
        return E_INVALIDARG;

    EnterCriticalSection (en->lock);

    if (en->enumerable)
        hr = IEnumerable_Skip (en->enumerable, &en->iterator, count);
    else
        hr = E_UNEXPECTED;

    LeaveCriticalSection (en->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CEnumerator_Reset (IEnumerator* This)
{
    CEnumerator* en;
    LPVOID newIterator;
    HRESULT hr;

    en = COM_GetObject (CEnumerator, This);
    if (!en)
        return E_INVALIDARG;

    EnterCriticalSection (en->lock);

    if (en->enumerable)
    {
        newIterator = NULL;

        hr = IEnumerable_CreateIterator (en->enumerable, &newIterator);
        if (SUCCEEDED (hr))
        {
            IEnumerable_FreeIterator (en->enumerable, en->iterator);
            en->iterator = newIterator;
        }
    }
    else
        hr = E_UNEXPECTED;

    LeaveCriticalSection (en->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static HRESULT CreateEnumeratorInternal (IEnumerable* container, IGuid* iid, LPVOID iterator, BOOL validIterator, IEnumerator** ppEnum)
{
    CEnumerator* en;
    HRESULT hr;

    if (!validIterator)
    {
        iterator = NULL;

        hr = IEnumerable_CreateIterator (container, &iterator);
        if (FAILED (hr))
            return hr;
    }

    hr = COM_CreateObject (&COM_CLASS_NAME(CEnumerator), NULL, &en); /* ref: 1 */
    if (FAILED (hr))
    {
        if (!validIterator)
            IEnumerable_FreeIterator (container, iterator);

        return hr;
    }

    IGuid_AddRef (iid);
    en->iid = iid;

    IEnumerable_AddRef (container);
    en->enumerable = container;

    en->iterator = iterator;

    hr = IUnknown_QueryInterface ((IUnknown*)en, IGuid_Value (iid), ppEnum); /* ref: 2 */
    IUnknown_Release ((IUnknown*)en); /* ref: 1 */

    if (hr != S_OK)
    {
        if (!validIterator)
            IEnumerable_FreeIterator (container, iterator);

        IEnumerable_Release (container);
        en->enumerable = NULL;

        IUnknown_Release ((IUnknown*)en); /* ref: 0 */

        return E_FAIL;
    }

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CEnumerator_Clone (IEnumerator* This, IEnumerator** ppEnum)
{
    CEnumerator* en;
    LPVOID newIterator;
    HRESULT hr;

    if (ppEnum)
        *ppEnum = NULL;
    else
        return E_POINTER;

    en = COM_GetObject (CEnumerator, This);
    if (!en)
        return E_INVALIDARG;

    EnterCriticalSection (en->lock);

    if (en->enumerable && en->iid)
    {
        newIterator = NULL;

        hr = IEnumerable_CopyIterator (en->enumerable, en->iterator, &newIterator);
        if (SUCCEEDED (hr))
        {
            hr = CreateEnumeratorInternal (en->enumerable, en->iid, newIterator, TRUE, ppEnum);
            if (FAILED (hr))
                IEnumerable_FreeIterator (en->enumerable, newIterator);
        }
    }
    else
        hr = E_UNEXPECTED;

    LeaveCriticalSection (en->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

HRESULT COM_CreateEnumerator (IEnumerable* container, REFIID enumIID, LPVOID* ppvObject)
{
    IGuid* iid;
    HRESULT hr;

    if (ppvObject)
        *ppvObject = NULL;
    else
        return E_POINTER;

    if (!container || !enumIID)
        return E_INVALIDARG;

    iid = COM_GetGuidInterface (enumIID);
    if (iid)
    {
        hr = CreateEnumeratorInternal (container, iid, NULL, FALSE, (IEnumerator**)ppvObject);

        IGuid_Release (iid);
    }
    else
        hr = E_OUTOFMEMORY;

    return hr;
}

/* -------------------------------------------------------------------------- */
/*                             CUnknArray                                */
/* -------------------------------------------------------------------------- */

/* {8C7A861B-D12E-406b-A445-F1EC93B54C89} */
static const GUID CLSID_CUnknArray = {0x8c7a861b, 0xd12e, 0x406b, {0xa4, 0x45, 0xf1, 0xec, 0x93, 0xb5, 0x4c, 0x89}};

/* IUnknArray */
static STDMETHODIMP CUnknArray_SetMaxSize   (IUnknArray* This, DWORD count);
static STDMETHODIMP CUnknArray_Add          (IUnknArray* This, IUnknown* unkn, REFIID iid/*=NULL*/, DWORD* pCookie);
static STDMETHODIMP CUnknArray_Remove       (IUnknArray* This, DWORD cookie);
static STDMETHODIMP CUnknArray_Clear        (IUnknArray* This);
static STDMETHODIMP CUnknArray_GetUnknown   (IUnknArray* This, DWORD dwCookie, REFIID iid, LPVOID* ppvObject);
static STDMETHODIMP CUnknArray_GetCount     (IUnknArray* This, DWORD* count);
static STDMETHODIMP CUnknArray_IsEmpty      (IUnknArray* This);

/* IEnumerable */
static STDMETHODIMP CUnknArray_CreateIterator   (IEnumerable* This, LPVOID* iter);
static STDMETHODIMP CUnknArray_CopyIterator     (IEnumerable* This, LPVOID iter, LPVOID* newIter);
static STDMETHODIMP CUnknArray_FreeIterator     (IEnumerable* This, LPVOID iter);
static STDMETHODIMP CUnknArray_Next             (IEnumerable* This, LPVOID* iter, ULONG count, LPVOID buffer, ULONG* pFetched/*=NULL*/);
static STDMETHODIMP CUnknArray_Skip             (IEnumerable* This, LPVOID* iter, ULONG count);

COM_DECLARE_VTABLE (CUnknArray, IUnknArray)
{
    /* IUnknown */
    COM_IUNKNOWN_METHODS (IUnknArray),

    /* IUnknArray */
    CUnknArray_SetMaxSize,
    CUnknArray_Add,
    CUnknArray_Remove,
    CUnknArray_Clear,
    CUnknArray_GetUnknown,
    CUnknArray_GetCount,
    CUnknArray_IsEmpty,
};

COM_DECLARE_VTABLE (CUnknArray, IEnumerable)
{
    /* IUnknown */
    COM_IUNKNOWN_METHODS (IEnumerable),

    /* IEnumerable */
    CUnknArray_CreateIterator,
    CUnknArray_CopyIterator,
    CUnknArray_FreeIterator,
    CUnknArray_Next,
    CUnknArray_Skip,
};

#define UNKNARRAY_MAXSIZE       (DWORD)(0xffffffff / sizeof(LPVOID))
#define UNKNARRAY_STATIC_ITEMS  4

typedef struct _CUnknArray
{
    COM_OBJECT object;

    /* implemented interfaces */
    IUnknArray* unknArray;
    IEnumerable* enumerable;

    /* aggregated interfaces */

    /* private data */
    LPCRITICAL_SECTION lock;

    union
    {
        IUnknown** unkn;
        DWORD* dw;
        IUnknown* unknStatic[UNKNARRAY_STATIC_ITEMS];
        DWORD dwStatic[UNKNARRAY_STATIC_ITEMS];
    }items;

    DWORD count;
    DWORD size;
    DWORD maxSize;

    IGlobalInterfaceTable* git;
}CUnknArray;

static HRESULT COM_CTOR_NAME (CUnknArray) (COM_OBJECT* object);
static void    COM_DTOR_NAME (CUnknArray) (COM_OBJECT* object);

BEGIN_INTERFACE_MAP(CUnknArray)
    COM_INTERFACE_ENTRY (IUnknArray, CUnknArray, unknArray),
    COM_INTERFACE_ENTRY (IEnumerable, CUnknArray, enumerable),
END_INTERFACE_MAP();

COM_DECLARE_PRIVATE_CLASS (NULL, CUnknArray, 0);

/* -------------------------------------------------------------------------- */

static HRESULT COM_CTOR_NAME (CUnknArray) (COM_OBJECT* object)
{
    CUnknArray* ua;

    ua = (CUnknArray*)object;

    ua->lock = COM_GetClassLock (object);
    if (!ua->lock)
    {
        ua->lock = _New (CRITICAL_SECTION);
        if (!ua->lock)
            return E_OUTOFMEMORY;

        InitializeCriticalSection (ua->lock);
    }

    ua->maxSize = UNKNARRAY_STATIC_ITEMS;

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static void COM_DTOR_NAME (CUnknArray) (COM_OBJECT* object)
{
    CUnknArray* ua;

    ua = (CUnknArray*)object;

    IUnknArray_Clear (ua->unknArray);

    if (ua->git)
    {
        IGlobalInterfaceTable_Release (ua->git);
        ua->git = NULL;
    }

    if (ua->lock && !COM_IsClassLock (object, ua->lock))
    {
        DeleteCriticalSection (ua->lock);
        _Delete (ua->lock, CRITICAL_SECTION);
        ua->lock = NULL;
    }
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_SetMaxSize (IUnknArray* This, DWORD count)
{
    CUnknArray* ua;
    HRESULT hr;

    COM_METHOD_PROLOGUE (ua, !ua || count == 0 || count > UNKNARRAY_MAXSIZE);

    EnterCriticalSection (ua->lock);
    
    if (count >= ua->size)
    {
        ua->maxSize = count;
        hr = S_OK;
    }
    else
        hr = E_INVALIDARG;

    LeaveCriticalSection (ua->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_Add (IUnknArray* This, IUnknown* pUnknown, REFIID iid/*=NULL*/, DWORD* pdwCookie)
{
    CUnknArray* ua;
    LPVOID items;
    DWORD dwCookie;
    DWORD i, size;
    HRESULT hr;

    COM_INIT_PTR_ARG_ (pdwCookie, 0);
    COM_METHOD_PROLOGUE (ua, !ua || !pUnknown || (ua->git && !iid));

    EnterCriticalSection (ua->lock);

    if (ua->count >= ua->maxSize)
    {
        LeaveCriticalSection (ua->lock);

        return CONNECT_E_ADVISELIMIT;
    }

    if (ua->size <= UNKNARRAY_STATIC_ITEMS)
    {
        /* SIZE <= UNKNARRAY_STATIC_ITEMS */

        if (ua->count < ua->size)
        {
            /* COUNT < SIZE */

            if (ua->git)
            {
                for (i=0; i<ua->size; i++)
                    if (ua->items.dwStatic[i] == 0)
                        break;

                if (i >= ua->size)
                {
                    LeaveCriticalSection (ua->lock);

                    return E_UNEXPECTED;
                }

                hr = IGlobalInterfaceTable_RegisterInterfaceInGlobal (ua->git, pUnknown, iid, &dwCookie);
                if (FAILED (hr))
                {
                    LeaveCriticalSection (ua->lock);

                    return hr;
                }

                ua->items.dwStatic[i] = dwCookie;
            }
            else
            {
                for (i=0; i<ua->size; i++)
                    if (ua->items.unknStatic[i] == NULL)
                        break;

                if (i >= ua->size)
                {
                    LeaveCriticalSection (ua->lock);

                    return E_UNEXPECTED;
                }

                IUnknown_AddRef (pUnknown);

                ua->items.unknStatic[i] = pUnknown;
            }

            dwCookie = i + 1;
        }
        else if (ua->count == ua->size)
        {
            /* COUNT == SIZE */

            size = MIN (ua->maxSize, UNKNARRAY_STATIC_ITEMS * 2);

            if (ua->git)
            {
                hr = IGlobalInterfaceTable_RegisterInterfaceInGlobal (ua->git, pUnknown, iid, &dwCookie);
                if (FAILED (hr))
                {
                    LeaveCriticalSection (ua->lock);

                    return hr;
                }

                items = _Alloc (size * sizeof(DWORD));
                if (!items)
                {
                    IGlobalInterfaceTable_RevokeInterfaceFromGlobal (ua->git, dwCookie);
                    LeaveCriticalSection (ua->lock);

                    return E_OUTOFMEMORY;
                }

                for (i=0; i<ua->size; i++)
                    ((DWORD*)items)[i] = ua->items.dwStatic[i];

                ua->items.dw = (DWORD*)items;
                ua->items.dw[ua->size] = dwCookie;
            }
            else
            {
                items = _Alloc (size * sizeof(IUnknown*));
                if (!items)
                {
                    LeaveCriticalSection (ua->lock);

                    return E_OUTOFMEMORY;
                }

                IUnknown_AddRef (pUnknown);

                for (i=0; i<ua->size; i++)
                    ((IUnknown**)items)[i] = ua->items.unknStatic[i];

                ua->items.unkn = (IUnknown**)items;
                ua->items.unkn[ua->size] = pUnknown;
            }

            dwCookie = ua->size + 1;
            ua->size = size;
        }
        else
        {
            /* COUNT > SIZE */

            LeaveCriticalSection (ua->lock);

            return E_UNEXPECTED;
        }
    }
    else
    {
        /* SIZE > UNKNARRAY_STATIC_ITEMS */

        if (ua->count == ua->size)
        {
            /* COUNT == SIZE */

            if (ua->size >= UNKNARRAY_MAXSIZE / 2)
            {
                LeaveCriticalSection (ua->lock);

                return CONNECT_E_ADVISELIMIT;
            }

            size = MIN (ua->maxSize, ua->size * 2);

            if (ua->git)
            {
                hr = IGlobalInterfaceTable_RegisterInterfaceInGlobal (ua->git, pUnknown, iid, &dwCookie);
                if (FAILED (hr))
                {
                    LeaveCriticalSection (ua->lock);

                    return hr;
                }

                items = _Realloc (ua->items.dw, ua->size * sizeof(DWORD), size * sizeof(DWORD));
                if (!items)
                {
                    IGlobalInterfaceTable_RevokeInterfaceFromGlobal (ua->git, dwCookie);
                    LeaveCriticalSection (ua->lock);

                    return E_OUTOFMEMORY;
                }

                ua->items.dw = (DWORD*)items;
                ua->items.dw[ua->size] = dwCookie;
            }
            else
            {
                items = _Realloc (ua->items.unkn, ua->size * sizeof(IUnknown*), size * sizeof(IUnknown*));
                if (!items)
                {
                    LeaveCriticalSection (ua->lock);

                    return E_OUTOFMEMORY;
                }

                IUnknown_AddRef (pUnknown);

                ua->items.unkn = (IUnknown**)items;
                ua->items.unkn[ua->size] = pUnknown;
            }

            dwCookie = ua->size + 1;
            ua->size = size;
        }
        else if (ua->count < ua->size)
        {
            /* COUNT < SIZE */

            if (ua->git)
            {
                for (i=0; i<ua->size; i++)
                {
                    if (ua->items.dw[i] == 0)
                    {
                        hr = IGlobalInterfaceTable_RegisterInterfaceInGlobal (ua->git, pUnknown, iid, &dwCookie);
                        if (FAILED (hr))
                        {
                            LeaveCriticalSection (ua->lock);

                            return hr;
                        }

                        ua->items.dw[i] = dwCookie;

                        break;
                    }
                }
            }
            else
            {
                for (i=0; i<ua->size; i++)
                {
                    if (ua->items.unkn[i] == NULL)
                    {
                        IUnknown_AddRef (pUnknown);
                        ua->items.unkn[i] = pUnknown;

                        break;
                    }
                }
            }

            if (i >= ua->size)
            {
                LeaveCriticalSection (ua->lock);

                return E_UNEXPECTED;
            }

            dwCookie = i + 1;
        }
        else
        {
            /* COUNT > SIZE*/

            LeaveCriticalSection (ua->lock);

            return E_UNEXPECTED;
        }
    }

    *pdwCookie = dwCookie;
    ua->count++;

    LeaveCriticalSection (ua->lock);

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_Remove (IUnknArray* This, DWORD dwCookie)
{
    CUnknArray* ua;
    IUnknown* unknStatic[UNKNARRAY_STATIC_ITEMS];
    DWORD dwStatic[UNKNARRAY_STATIC_ITEMS];
    DWORD i, count;

    COM_METHOD_PROLOGUE (ua, !ua || dwCookie == 0);

    EnterCriticalSection (ua->lock);

    dwCookie--;

    if (dwCookie >= ua->size || ua->count == 0)
    {
        LeaveCriticalSection (ua->lock);

        return CONNECT_E_NOCONNECTION;
    }

    if (ua->size <= UNKNARRAY_STATIC_ITEMS)
    {
        /* SIZE <= UNKNARRAY_STATIC_ITEMS */

        if (ua->git)
        {
            if (ua->items.dwStatic[dwCookie] == 0)
            {
                LeaveCriticalSection (ua->lock);

                return CONNECT_E_NOCONNECTION;
            }

            IGlobalInterfaceTable_RevokeInterfaceFromGlobal (ua->git, ua->items.dwStatic[dwCookie]);
            ua->items.dwStatic[dwCookie] = 0;
        }
        else
        {
            if (ua->items.unknStatic[dwCookie] == NULL)
            {
                LeaveCriticalSection (ua->lock);

                return CONNECT_E_NOCONNECTION;
            }

            IUnknown_Release (ua->items.unknStatic[dwCookie]);
            ua->items.unknStatic[dwCookie] = NULL;
        }
    }
    else
    {
        /* SIZE > UNKNARRAY_STATIC_ITEMS */

        if (ua->git)
        {
            if (ua->items.dw[dwCookie] == 0)
            {
                LeaveCriticalSection (ua->lock);

                return CONNECT_E_NOCONNECTION;
            }

            IGlobalInterfaceTable_RevokeInterfaceFromGlobal (ua->git, ua->items.dw[dwCookie]);
            ua->items.dw[dwCookie] = 0;
        }
        else
        {
            if (ua->items.unkn[dwCookie] == NULL)
            {
                LeaveCriticalSection (ua->lock);

                return CONNECT_E_NOCONNECTION;
            }

            IUnknown_Release (ua->items.unkn[dwCookie]);
            ua->items.unkn[dwCookie] = NULL;
        }

        if ((ua->count - 1) <= UNKNARRAY_STATIC_ITEMS)
        {
            if (ua->git)
            {
                for (i=0, count=0; i<UNKNARRAY_STATIC_ITEMS; i++)
                    if (ua->items.dw[i] != 0)
                        count++;

                if (count == ua->count - 1)
                {
                    for (i=0; i<UNKNARRAY_STATIC_ITEMS; i++)
                        dwStatic[i] = ua->items.dw[i];

                    _Free (ua->items.dw, ua->size * sizeof(DWORD));

                    for (i=0; i<UNKNARRAY_STATIC_ITEMS; i++)
                        ua->items.dwStatic[i] = dwStatic[i];

                    ua->size = MIN (ua->maxSize, UNKNARRAY_STATIC_ITEMS);
                }
            }
            else
            {
                for (i=0, count=0; i<UNKNARRAY_STATIC_ITEMS; i++)
                    if (ua->items.unkn[i] != 0)
                        count++;

                if (count == ua->count - 1)
                {
                    for (i=0; i<UNKNARRAY_STATIC_ITEMS; i++)
                        unknStatic[i] = ua->items.unkn[i];

                    _Free (ua->items.unkn, ua->size * sizeof(IUnknown*));

                    for (i=0; i<UNKNARRAY_STATIC_ITEMS; i++)
                        ua->items.unknStatic[i] = unknStatic[i];

                    ua->size = MIN (ua->maxSize, UNKNARRAY_STATIC_ITEMS);
                }
            }

        }
    }

    ua->count--;

    LeaveCriticalSection (ua->lock);

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_Clear (IUnknArray* This)
{
    CUnknArray* ua;
    DWORD i;

    COM_METHOD_PROLOGUE (ua, !ua);

    EnterCriticalSection (ua->lock);

    if (ua->git)
    {
        if (ua->size <= UNKNARRAY_STATIC_ITEMS)
        {
            for (i=0; i<ua->size; i++)
            {
                if (ua->items.dwStatic[i])
                {
                    IGlobalInterfaceTable_RevokeInterfaceFromGlobal (ua->git, ua->items.dwStatic[i]);
                    ua->items.dwStatic[i] = 0;
                }
            }
        }
        else
        {
            if (ua->items.dw)
            {
                for (i=0; i<ua->size; i++)
                {
                    if (ua->items.dw[i])
                        IGlobalInterfaceTable_RevokeInterfaceFromGlobal (ua->git, ua->items.dw[i]);
                }

                _Free (ua->items.dw, ua->size * sizeof(DWORD));

                for (i=0; i<UNKNARRAY_STATIC_ITEMS; i++)
                    ua->items.dwStatic[i] = 0;
            }
        }
    }
    else
    {
        if (ua->size <= UNKNARRAY_STATIC_ITEMS)
        {
            for (i=0; i<ua->size; i++)
            {
                if (ua->items.unknStatic[i])
                {
                    IUnknown_Release (ua->items.unknStatic[i]);
                    ua->items.unknStatic[i] = NULL;
                }
            }
        }
        else
        {
            if (ua->items.unkn)
            {
                for (i=0; i<ua->size; i++)
                    if (ua->items.unkn[i])
                        IUnknown_Release (ua->items.unkn[i]);

                _Free (ua->items.unkn, ua->size * sizeof(IUnknown*));

                for (i=0; i<UNKNARRAY_STATIC_ITEMS; i++)
                    ua->items.unknStatic[i] = NULL;
            }
        }
    }

    ua->count = 0;
    ua->size = MIN (ua->maxSize, UNKNARRAY_STATIC_ITEMS);

    LeaveCriticalSection (ua->lock);

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_GetUnknown (IUnknArray* This, DWORD dwCookie, REFIID iid, LPVOID* ppvObject)
{
    CUnknArray* ua;
    HRESULT hr;

    COM_INIT_PTR_ARG (ppvObject);
    COM_METHOD_PROLOGUE (ua, !ua || dwCookie == 0 || (ua->git && !iid));

    EnterCriticalSection (ua->lock);

    dwCookie--;

    if (dwCookie >= ua->size || ua->count == 0)
    {
        LeaveCriticalSection (ua->lock);

        return CONNECT_E_NOCONNECTION;
    }

    if (ua->size <= UNKNARRAY_STATIC_ITEMS)
    {
        /* SIZE <= UNKNARRAY_STATIC_ITEMS */

        if (ua->git)
        {
            if (ua->items.dwStatic[dwCookie] == 0)
            {
                LeaveCriticalSection (ua->lock);

                return CONNECT_E_NOCONNECTION;
            }

            hr = IGlobalInterfaceTable_GetInterfaceFromGlobal (ua->git, ua->items.dwStatic[dwCookie], iid, ppvObject);
        }
        else
        {
            if (ua->items.unknStatic[dwCookie] == NULL)
            {
                LeaveCriticalSection (ua->lock);

                return CONNECT_E_NOCONNECTION;
            }

            if (!iid)
            {
                *ppvObject = ua->items.unknStatic[dwCookie];
                IUnknown_AddRef ((IUnknown*)*ppvObject);
                hr = S_OK;
            }
            else
                hr = IUnknown_QueryInterface (ua->items.unknStatic[dwCookie], iid, ppvObject);
        }
    }
    else
    {
        /* SIZE > UNKNARRAY_STATIC_ITEMS */

        if (ua->git)
        {
            if (ua->items.dw[dwCookie] == 0)
            {
                LeaveCriticalSection (ua->lock);

                return CONNECT_E_NOCONNECTION;
            }

            hr = IGlobalInterfaceTable_GetInterfaceFromGlobal (ua->git, ua->items.dw[dwCookie], iid, ppvObject);
        }
        else
        {
            if (ua->items.unkn[dwCookie] == NULL)
            {
                LeaveCriticalSection (ua->lock);

                return CONNECT_E_NOCONNECTION;
            }

            if (!iid)
            {
                *ppvObject = ua->items.unkn[dwCookie];
                IUnknown_AddRef ((IUnknown*)*ppvObject);
                hr = S_OK;
            }
            else
                hr = IUnknown_QueryInterface (ua->items.unkn[dwCookie], iid, ppvObject);
        }
    }

    LeaveCriticalSection (ua->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_GetCount (IUnknArray* This, DWORD* count)
{
    CUnknArray* ua;

    COM_INIT_PTR_ARG_ (count, 0);
    COM_METHOD_PROLOGUE (ua, !ua);

    *count = ua->count;

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_IsEmpty (IUnknArray* This)
{
    CUnknArray* ua;

    ua = COM_GetObject (CUnknArray, This);

    return ((ua && ua->count != 0) ? S_FALSE : S_OK);
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_CreateIterator (IEnumerable* This, LPVOID* iter)
{
    COM_INIT_PTR_ARG_ (iter, (LPVOID)0);

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_CopyIterator (IEnumerable* This, LPVOID iter, LPVOID* newIter)
{
    COM_INIT_PTR_ARG_ (newIter, iter);

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_FreeIterator (IEnumerable* This, LPVOID iter)
{
    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_Next (IEnumerable* This, LPVOID* iter, ULONG count, LPVOID buffer, ULONG* pFetched/*=NULL*/)
{
    CUnknArray* ua;
    CONNECTDATA* cd;
    DWORD i, copied;
    HRESULT hr;

    if (pFetched)
        *pFetched = 0;

    COM_METHOD_PROLOGUE (ua, !ua || !iter || (count && !buffer));

    if (count == 0)
        return S_OK;

    EnterCriticalSection (ua->lock);

    copied = 0;
    cd = (CONNECTDATA*)buffer;

    if (ua->git)
    {
        if (ua->size <= UNKNARRAY_STATIC_ITEMS)
        {
            for (i=*(DWORD*)iter; i<ua->size; i++)
            {
                if (ua->items.dwStatic[i])
                {
                    hr = IGlobalInterfaceTable_GetInterfaceFromGlobal (ua->git, ua->items.dwStatic[i], &IID_IUnknown, &cd->pUnk);
                    if (SUCCEEDED (hr))
                    {
                        cd->dwCookie = i + 1;
                        cd++;
                        copied++;

                        if (copied >= count)
                        {
                            i++;
                            break;
                        }
                    }
                }
            }
        }
        else
        {
            for (i=*(DWORD*)iter; i<ua->size; i++)
            {
                if (ua->items.dw[i])
                {
                    hr = IGlobalInterfaceTable_GetInterfaceFromGlobal (ua->git, ua->items.dw[i], &IID_IUnknown, &cd->pUnk);
                    if (SUCCEEDED (hr))
                    {
                        cd->dwCookie = i + 1;
                        cd++;
                        copied++;

                        if (copied >= count)
                        {
                            i++;
                            break;
                        }
                    }
                }
            }
        }
    }
    else
    {
        if (ua->size <= UNKNARRAY_STATIC_ITEMS)
        {
            for (i=*(DWORD*)iter; i<ua->size; i++)
            {
                if (ua->items.unknStatic[i])
                {
                    cd->pUnk = ua->items.unknStatic[i];
                    IUnknown_AddRef (cd->pUnk);
                    cd->dwCookie = i + 1;
                    cd++;
                    copied++;

                    if (copied >= count)
                    {
                        i++;
                        break;
                    }
                }
            }
        }
        else
        {
            for (i=*(DWORD*)iter; i<ua->size; i++)
            {
                if (ua->items.unkn[i])
                {
                    cd->pUnk = ua->items.unkn[i];
                    IUnknown_AddRef (cd->pUnk);
                    cd->dwCookie = i + 1;
                    cd++;
                    copied++;

                    if (copied >= count)
                    {
                        i++;
                        break;
                    }
                }
            }
        }
    }

    *(DWORD*)iter = i;

    LeaveCriticalSection (ua->lock);

    if (pFetched)
        *pFetched = copied;

    return ((copied == count) ? S_OK : S_FALSE);
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CUnknArray_Skip (IEnumerable* This, LPVOID* iter, ULONG count)
{
    CUnknArray* ua;
    DWORD i, skipped;

    COM_METHOD_PROLOGUE (ua, !ua || !iter);

    if (count == 0)
        return S_OK;

    EnterCriticalSection (ua->lock);

    skipped = 0;

    if (ua->git)
    {
        if (ua->size <= UNKNARRAY_STATIC_ITEMS)
        {
            for (i=*(DWORD*)iter; i<ua->size; i++)
            {
                if (ua->items.dwStatic[i])
                {
                    skipped++;

                    if (skipped >= count)
                    {
                        i++;
                        break;
                    }
                }
            }
        }
        else
        {
            for (i=*(DWORD*)iter; i<ua->size; i++)
            {
                if (ua->items.dw[i])
                {
                    skipped++;

                    if (skipped >= count)
                    {
                        i++;
                        break;
                    }
                }
            }
        }
    }
    else
    {
        if (ua->size <= UNKNARRAY_STATIC_ITEMS)
        {
            for (i=*(DWORD*)iter; i<ua->size; i++)
            {
                if (ua->items.unknStatic[i])
                {
                    skipped++;

                    if (skipped >= count)
                    {
                        i++;
                        break;
                    }
                }
            }
        }
        else
        {
            for (i=*(DWORD*)iter; i<ua->size; i++)
            {
                if (ua->items.unkn[i])
                {
                    skipped++;

                    if (skipped >= count)
                    {
                        i++;
                        break;
                    }
                }
            }
        }
    }

    *(DWORD*)iter = i;
    
    LeaveCriticalSection (ua->lock);

    return ((skipped == count) ? S_OK : S_FALSE);
}

/* -------------------------------------------------------------------------- */

HRESULT COM_CreateUnknArray (DWORD maxSize, IUnknArray** ppUnknArray)
{
    CUnknArray* ua;
    HRESULT hr;

    COM_INIT_PTR_ARG (ppUnknArray);

    if (maxSize == 0)
        return E_INVALIDARG;

    if (maxSize > UNKNARRAY_MAXSIZE)
        maxSize = UNKNARRAY_MAXSIZE;

    hr = COM_CreateObject (&COM_CLASS_NAME (CUnknArray), NULL, &ua);
    if (SUCCEEDED (hr))
    {
        ua->maxSize = maxSize;
        *ppUnknArray = ua->unknArray;
    }

    return hr;
}

/* -------------------------------------------------------------------------- */

HRESULT COM_CreateGitUnknArray (DWORD maxSize, IUnknArray** ppUnknArray)
{
    CUnknArray* ua;
    HRESULT hr;

    COM_INIT_PTR_ARG (ppUnknArray);

    if (maxSize == 0)
        return E_INVALIDARG;

    if (maxSize > UNKNARRAY_MAXSIZE)
        maxSize = UNKNARRAY_MAXSIZE;

    hr = COM_CreateObject (&COM_CLASS_NAME (CUnknArray), NULL, &ua);
    if (SUCCEEDED (hr))
    {
        ua->maxSize = maxSize;
        hr = COM_GetGIT (&ua->git);

        if (SUCCEEDED (hr))
            *ppUnknArray = ua->unknArray;
        else
            IUnknArray_Release (ua->unknArray);
    }

    return hr;
}

/* -------------------------------------------------------------------------- */
/*                        CConnectionPoint                               */
/* -------------------------------------------------------------------------- */

/* {368032F4-88A5-4648-9C4F-893C5F583091} */
static const GUID CLSID_CConnectionPoint = {0x368032f4, 0x88a5, 0x4648, {0x9c, 0x4f, 0x89, 0x3c, 0x5f, 0x58, 0x30, 0x91}};

static STDMETHODIMP CConnectionPoint_GetConnectionInterface       (IConnectionPoint* This, IID* pIID);
static STDMETHODIMP CConnectionPoint_GetConnectionPointContainer  (IConnectionPoint* This, IConnectionPointContainer** ppCPC);
static STDMETHODIMP CConnectionPoint_Advise                       (IConnectionPoint* This, IUnknown* pUnkSink, DWORD* pdwCookie);
static STDMETHODIMP CConnectionPoint_Unadvise                     (IConnectionPoint* This, DWORD dwCookie);
static STDMETHODIMP CConnectionPoint_EnumConnections              (IConnectionPoint* This, IEnumConnections **ppEnum);

COM_DECLARE_VTABLE(CConnectionPoint, IConnectionPoint)
{
    /* IUnknown */
    COM_IUNKNOWN_METHODS (IConnectionPoint),

    /* IConnectionPoint */
    CConnectionPoint_GetConnectionInterface,
    CConnectionPoint_GetConnectionPointContainer,
    CConnectionPoint_Advise,
    CConnectionPoint_Unadvise,
    CConnectionPoint_EnumConnections,
};

typedef struct _CConnectionPoint
{
    COM_OBJECT object;

    /* implemented interfaces */
    IConnectionPoint* intf;

    /* aggregated interfaces */

    /* private data */
    IGuid* iid;

    IConnectionPointContainer* container;

    CRITICAL_SECTION* lock;
    IUnknArray* unknArray;
    DWORD maxConnections;
}CConnectionPoint;

static HRESULT COM_CTOR_NAME (CConnectionPoint) (COM_OBJECT* object);
static void    COM_DTOR_NAME (CConnectionPoint) (COM_OBJECT* object);

BEGIN_INTERFACE_MAP(CConnectionPoint)
    COM_INTERFACE_ENTRY (IConnectionPoint, CConnectionPoint, intf),
END_INTERFACE_MAP();

COM_DECLARE_PRIVATE_CLASS (NULL, CConnectionPoint, 0);

/* -------------------------------------------------------------------------- */

static HRESULT COM_CTOR_NAME (CConnectionPoint) (COM_OBJECT* object)
{
    CConnectionPoint* cp;

    cp = (CConnectionPoint*)object;

    cp->lock = COM_GetClassLock (object);
    if (!cp->lock)
    {
        cp->lock = _New (CRITICAL_SECTION);
        if (!cp->lock)
            return E_OUTOFMEMORY;

        InitializeCriticalSection (cp->lock);
    }

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static void COM_DTOR_NAME (CConnectionPoint) (COM_OBJECT* object)
{
    CConnectionPoint* cp;

    cp = (CConnectionPoint*)object;

    if (cp->iid)
    {
        IGuid_Release (cp->iid);
        cp->iid = NULL;
    }

    if (cp->unknArray)
    {
        IUnknArray_Release (cp->unknArray);
        cp->unknArray = NULL;
    }

    if (cp->container)
    {
/*
        IConnectionPointContainer_Release (cp->container);
*/
        cp->container = NULL;
    }

    if (!COM_IsClassLock (object, cp->lock))
    {
        DeleteCriticalSection (cp->lock);
        _Delete (cp->lock, CRITICAL_SECTION);
    }
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectionPoint_GetConnectionInterface (IConnectionPoint* This, IID* pIID)
{
    CConnectionPoint* cp;

    COM_METHOD_PROLOGUE (cp, !cp || !pIID);

    EnterCriticalSection (cp->lock);

    if (!cp->iid)
    {
        LeaveCriticalSection (cp->lock);

        return E_UNEXPECTED;
    }

    COM_CopyGUID (pIID, IGuid_Value (cp->iid));

    LeaveCriticalSection (cp->lock);

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectionPoint_GetConnectionPointContainer (IConnectionPoint* This, IConnectionPointContainer** ppCPC)
{
    CConnectionPoint* cp;

    COM_INIT_PTR_ARG (ppCPC);
    COM_METHOD_PROLOGUE (cp, !cp);

    EnterCriticalSection (cp->lock);

    if (!cp->container)
    {
        LeaveCriticalSection (cp->lock);

        return E_UNEXPECTED;
    }

    IConnectionPointContainer_AddRef (cp->container);
    *ppCPC = cp->container;

    LeaveCriticalSection (cp->lock);

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectionPoint_Advise (IConnectionPoint* This, IUnknown* pUnkSink, DWORD* pdwCookie)
{
    CConnectionPoint* cp;
    LPCGUID iid;
    IUnknown* intf;
    DWORD dwCookie;
    HRESULT hr;

    COM_INIT_PTR_ARG_ (pdwCookie, 0);
    COM_METHOD_PROLOGUE (cp, !cp || !pUnkSink);

    EnterCriticalSection (cp->lock);

    if (cp->iid)
    {
        iid = IGuid_Value (cp->iid);

        if (cp->maxConnections != 0)
        {
            hr = IUnknown_QueryInterface (pUnkSink, iid, &intf);
            if (SUCCEEDED (hr))
            {
                if (!cp->unknArray)
                    hr = COM_CreateGitUnknArray (cp->maxConnections, &cp->unknArray);

                if (cp->unknArray)
                {
                    hr = IUnknArray_Add (cp->unknArray, intf, iid, &dwCookie);
                    if (FAILED (hr))
                    {
                        if (IUnknArray_IsEmpty (cp->unknArray) == S_OK)
                        {
                            IUnknArray_Release (cp->unknArray);
                            cp->unknArray = NULL;
                        }
                    }
                    else
                        *pdwCookie = dwCookie;
                }

                IUnknown_Release (intf);
            }
            else if (hr == E_NOINTERFACE)
                hr = CONNECT_E_CANNOTCONNECT;
        }
        else
            hr = CONNECT_E_ADVISELIMIT;
    }
    else
        hr = E_UNEXPECTED;

    LeaveCriticalSection (cp->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectionPoint_Unadvise (IConnectionPoint* This, DWORD dwCookie)
{
    CConnectionPoint* cp;
    HRESULT hr;

    COM_METHOD_PROLOGUE (cp, !cp);

    EnterCriticalSection (cp->lock);

    if (cp->unknArray)
    {
        hr = IUnknArray_Remove (cp->unknArray, dwCookie);

        if (IUnknArray_IsEmpty (cp->unknArray) == S_OK)
        {
            IUnknArray_Release (cp->unknArray);
            cp->unknArray = NULL;
        }
    }
    else
        hr = CONNECT_E_NOCONNECTION;

    LeaveCriticalSection (cp->lock);

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectionPoint_EnumConnections (IConnectionPoint* This, IEnumConnections **ppEnum)
{
    CConnectionPoint* cp;
    IEnumerable* enumerable;
    HRESULT hr;

    COM_INIT_PTR_ARG (ppEnum);
    COM_METHOD_PROLOGUE (cp, !cp);

    EnterCriticalSection (cp->lock);

    if (cp->unknArray)
    {
        hr = IUnknArray_QueryInterface (cp->unknArray, &IID_IEnumerable, &enumerable);
        if (SUCCEEDED (hr))
        {
            hr = COM_CreateEnumerator (enumerable, &IID_IEnumConnections, ppEnum);
            IEnumerable_Release (enumerable);
        }
    }
    else
        hr = E_EMPTY;

    LeaveCriticalSection (cp->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */
/*                          CConnectable                                 */
/* -------------------------------------------------------------------------- */

/* {7C274D4B-6311-4900-8EDB-61499317C06F} */
static const GUID CLSID_CConnectable = {0x7c274d4b, 0x6311, 0x4900, {0x8e, 0xdb, 0x61, 0x49, 0x93, 0x17, 0xc0, 0x6f}};

/* IConnectionPointContainer */
static STDMETHODIMP CConnectable_EnumConnectionPoints   (IConnectionPointContainer* This, IEnumConnectionPoints** ppEnum);
static STDMETHODIMP CConnectable_FindConnectionPoint    (IConnectionPointContainer* This, REFIID riid, IConnectionPoint** ppCP);

/* IEnumerable */
static STDMETHODIMP CConnectable_CreateIterator (IEnumerable* This, LPVOID* iter);
static STDMETHODIMP CConnectable_CopyIterator   (IEnumerable* This, LPVOID iter, LPVOID* newIter);
static STDMETHODIMP CConnectable_FreeIterator   (IEnumerable* This, LPVOID iter);
static STDMETHODIMP CConnectable_Next           (IEnumerable* This, LPVOID* iter, ULONG count, LPVOID buffer, ULONG* pFetched/*=NULL*/);
static STDMETHODIMP CConnectable_Skip           (IEnumerable* This, LPVOID* iter, ULONG count);

COM_DECLARE_VTABLE(CConnectable, IConnectionPointContainer)
{
    /* IUnknown */
    COM_IUNKNOWN_METHODS (IConnectionPointContainer),

    /* IConnectionPointContainer */
    CConnectable_EnumConnectionPoints,
    CConnectable_FindConnectionPoint,
};

COM_DECLARE_VTABLE (CConnectable, IEnumerable)
{
    /* IUnknown */
    COM_IUNKNOWN_METHODS (IEnumerable),

    /* IEnumerable */
    CConnectable_CreateIterator,
    CConnectable_CopyIterator,
    CConnectable_FreeIterator,
    CConnectable_Next,
    CConnectable_Skip,
};

/*   CConnectable::connectionPoints */
typedef struct _CCP_NODE
{
    COM_TREE_NODE node;

    CConnectionPoint* cp;
}CCP_NODE;

static HRESULT COM_CTOR_NAME (CConnectable) (COM_OBJECT* object);
static void    COM_DTOR_NAME (CConnectable) (COM_OBJECT* object);

BEGIN_INTERFACE_MAP(CConnectable)
    COM_INTERFACE_ENTRY (IConnectionPointContainer, CConnectable, container),
    COM_INTERFACE_ENTRY (IEnumerable, CConnectable, enumerable),
END_INTERFACE_MAP();

COM_DECLARE_PUBLIC_CLASS_NAME (NULL, "CConnectable<COM_CLASS>", CConnectable, 0);

static HRESULT CConnectable_CreateConnectionPoint (CConnectable* This, REFIID iid, DWORD maxConnections);

/* -------------------------------------------------------------------------- */

static void FreeConnectionPointNode (COM_TREE* tree, COM_TREE_NODE* node)
{
    CCP_NODE* ccpNode;

    ccpNode = (CCP_NODE*)node;
    if (ccpNode)
    {
        if (ccpNode->cp)
        {
            EnterCriticalSection (ccpNode->cp->lock);

            ccpNode->cp->container = NULL;

            LeaveCriticalSection (ccpNode->cp->lock);

            IConnectionPoint_Release (ccpNode->cp->intf);
        }

        _Delete (ccpNode, CCP_NODE);
    }
}

/* -------------------------------------------------------------------------- */

static int CompareConnectionPointNodes (COM_TREE* tree, LPCVOID key, COM_TREE_NODE* node)
{
    CCP_NODE* ccpNode;

    ccpNode = (CCP_NODE*)node;

    if (ccpNode && ccpNode->cp->iid && key)
        return IGuid_Compare (ccpNode->cp->iid, (REFGUID)key);

    return -1;
}

/* -------------------------------------------------------------------------- */

static HRESULT COM_CTOR_NAME (CConnectable) (COM_OBJECT* object)
{
    CConnectable* This;

    This = (CConnectable*)object;

    This->lock = COM_GetClassLock (object);
    if (!This->lock)
    {
        This->lock = _New (CRITICAL_SECTION);
        if (!This->lock)
            return E_OUTOFMEMORY;

        InitializeCriticalSection (This->lock);
    }

    This->connectionPoints.Free = FreeConnectionPointNode;
    This->connectionPoints.Compare = CompareConnectionPointNodes;

    This->CreateConnectionPoint = CConnectable_CreateConnectionPoint;

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static void COM_DTOR_NAME (CConnectable) (COM_OBJECT* object)
{
    CConnectable* This;

    This = (CConnectable*)object;

    COM_ClearTree (&This->connectionPoints);

    if (!COM_IsClassLock (object, This->lock))
    {
        DeleteCriticalSection (This->lock);
        _Delete (This->lock, CRITICAL_SECTION);
    }
}

/* -------------------------------------------------------------------------- */

static HRESULT CConnectable_CreateConnectionPoint (CConnectable* This, REFIID iid, DWORD maxConnections)
{
    CCP_NODE* ccpNode;
    HRESULT hr;

    if (!This || !iid || maxConnections == 0)
        return E_INVALIDARG;

    EnterCriticalSection (This->lock);

    if (!COM_FindNode (&This->connectionPoints, iid, NULL))
    {
        ccpNode = _New (CCP_NODE);
        if (ccpNode)
        {
            hr = COM_CreateObject (&COM_CLASS_NAME (CConnectionPoint), NULL, &ccpNode->cp);
            if (SUCCEEDED (hr))
            {
                ccpNode->cp->container = This->container;
                ccpNode->cp->maxConnections = maxConnections;

                ccpNode->cp->iid = COM_GetGuidInterface (iid);
                if (ccpNode->cp->iid)
                {
                    if (SUCCEEDED (hr))
                        hr = COM_AddNode (&This->connectionPoints, iid, &ccpNode->node) ? S_OK : E_UNEXPECTED;
                }
                else
                    hr = E_OUTOFMEMORY;
            }

            if (hr != S_OK)
                FreeConnectionPointNode (NULL, &ccpNode->node);
        }
        else
            hr = E_OUTOFMEMORY;
    }
    else
        hr = E_ALREADYEXISTS;

    LeaveCriticalSection (This->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectable_EnumConnectionPoints (IConnectionPointContainer* This, IEnumConnectionPoints** ppEnum)
{
    CConnectable* connectable;
    HRESULT hr;

    COM_INIT_PTR_ARG (ppEnum);
    COM_METHOD_PROLOGUE (connectable, !connectable);

    EnterCriticalSection (connectable->lock);
    
    hr = COM_CreateEnumerator (connectable->enumerable, &IID_IEnumConnectionPoints, ppEnum);

    LeaveCriticalSection (connectable->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectable_FindConnectionPoint (IConnectionPointContainer* This, REFIID riid, IConnectionPoint** ppCP)
{
    CConnectable* connectable;
    CCP_NODE* ccpNode;
    HRESULT hr;

    COM_INIT_PTR_ARG (ppCP);
    COM_METHOD_PROLOGUE (connectable, !connectable || !riid);

    EnterCriticalSection (connectable->lock);

    ccpNode = (CCP_NODE*)COM_FindNode (&connectable->connectionPoints, riid, NULL);
    if (ccpNode && ccpNode->cp)
    {
        IConnectionPoint_AddRef (ccpNode->cp->intf);
        *ppCP = ccpNode->cp->intf;
        hr = S_OK;
    }
    else
        hr = CONNECT_E_NOCONNECTION;

    LeaveCriticalSection (connectable->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectable_CreateIterator (IEnumerable* This, LPVOID* iter)
{
    CConnectable* connectable;
    HRESULT hr;

    COM_INIT_PTR_ARG (iter);
    COM_METHOD_PROLOGUE (connectable, !connectable);

    EnterCriticalSection (connectable->lock);

    *iter = COM_CreateTreeIterator (&connectable->connectionPoints);
    hr = (*iter) ? S_OK : E_OUTOFMEMORY;

    LeaveCriticalSection (connectable->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectable_CopyIterator (IEnumerable* This, LPVOID iter, LPVOID* newIter)
{
    CConnectable* connectable;
    HRESULT hr;

    COM_INIT_PTR_ARG (newIter);
    COM_METHOD_PROLOGUE (connectable, !connectable || !iter);

    EnterCriticalSection (connectable->lock);

    *newIter = COM_CopyTreeIterator (&connectable->connectionPoints, (COM_TREE_ITER*)iter);
    hr = (*newIter) ? S_OK : E_OUTOFMEMORY;

    LeaveCriticalSection (connectable->lock);

    return hr;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectable_FreeIterator (IEnumerable* This, LPVOID iter)
{
    CConnectable* connectable;

    COM_METHOD_PROLOGUE (connectable, !connectable || !iter);

    EnterCriticalSection (connectable->lock);

    COM_FreeTreeIterator (&connectable->connectionPoints, (COM_TREE_ITER*)iter);

    LeaveCriticalSection (connectable->lock);

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectable_Next (IEnumerable* This, LPVOID* iter, ULONG count, LPVOID buffer, ULONG* pFetched/*=NULL*/)
{
    CConnectable* connectable;
    COM_TREE_ITER* treeIter;
    IConnectionPoint** cp;
    DWORD copied;

    if (pFetched)
        *pFetched = 0;

    if (count == 0)
        return S_OK;

    COM_METHOD_PROLOGUE (connectable, !connectable || !iter || (count && !buffer));

    EnterCriticalSection (connectable->lock);
    
    treeIter = (COM_TREE_ITER*)(*iter);

    if (!COM_IsValidTreeIterator (&connectable->connectionPoints, treeIter))
    {
        LeaveCriticalSection (connectable->lock);

        return E_INVALIDARG;
    }

    copied = 0;
    cp = (IConnectionPoint**)buffer;

    if (treeIter->node == COM_TREE_ITER_FIRST_NODE)
        treeIter->node = COM_FirstNode (&connectable->connectionPoints);

    while (treeIter->node && copied < count)
    {
        if (((CCP_NODE*)treeIter->node)->cp)
        {
            *cp = ((CCP_NODE*)treeIter->node)->cp->intf;
            IConnectionPoint_AddRef (*cp);
            cp++;
            copied++;
        }

        treeIter->node = COM_NextNode (&connectable->connectionPoints, treeIter->node);
    }

    LeaveCriticalSection (connectable->lock);

    if (pFetched)
        *pFetched = copied;

    return ((copied == count) ? S_OK : S_FALSE);
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CConnectable_Skip (IEnumerable* This, LPVOID* iter, ULONG count)
{
    CConnectable* connectable;
    COM_TREE_ITER* treeIter;
    DWORD skipped;

    if (count == 0)
        return S_OK;

    COM_METHOD_PROLOGUE (connectable, !connectable || !iter);

    EnterCriticalSection (connectable->lock);

    treeIter = (COM_TREE_ITER*)(*iter);

    if (!COM_IsValidTreeIterator (&connectable->connectionPoints, treeIter))
    {
        LeaveCriticalSection (connectable->lock);

        return E_INVALIDARG;
    }

    skipped = 0;

    if (treeIter->node == COM_TREE_ITER_FIRST_NODE)
        treeIter->node = COM_FirstNode (&connectable->connectionPoints);

    while (treeIter->node && skipped < count)
    {
        if (((CCP_NODE*)treeIter->node)->cp)
            skipped++;

        treeIter->node = COM_NextNode (&connectable->connectionPoints, treeIter->node);
    }

    LeaveCriticalSection (connectable->lock);

    return ((skipped == count) ? S_OK : S_FALSE);
}

/* -------------------------------------------------------------------------- */
/*                       CClassFactory<COM_CLASS>                        */
/* -------------------------------------------------------------------------- */

/* {3545B1BE-7FD3-4f12-B709-4C324319671A} */
static const GUID CLSID_CClassFactory = {0x3545b1be, 0x7fd3, 0x4f12, {0xb7, 0x9, 0x4c, 0x32, 0x43, 0x19, 0x67, 0x1a}};

static STDMETHODIMP CClassFactory_CreateInstance  (IClassFactory* This, IUnknown *pUnkOuter, REFIID riid, void **ppvObject);
static STDMETHODIMP CClassFactory_LockServer      (IClassFactory* This, BOOL fLock);

COM_DECLARE_VTABLE (CClassFactory, IClassFactory)
{
    /* IUnknown */
    COM_IUNKNOWN_METHODS (IClassFactory),

    /* IClassFactory */
    CClassFactory_CreateInstance,
    CClassFactory_LockServer,
};

typedef struct _CClassFactory
{
    COM_OBJECT object;

    /* implemented interfaces */
    IClassFactory* intfClassFactory;

    /* aggregated interfaces */

    /* private members */
    const COM_CLASS* objectClass;
}CClassFactory;

static HRESULT COM_CTOR_NAME (CClassFactory) (COM_OBJECT* object);
static void    COM_DTOR_NAME (CClassFactory) (COM_OBJECT* object);

BEGIN_INTERFACE_MAP (CClassFactory)
    COM_INTERFACE_ENTRY (IClassFactory, CClassFactory, intfClassFactory),
END_INTERFACE_MAP();

COM_DECLARE_PRIVATE_CLASS_NAME (NULL, "CClassFactory<COM_CLASS>", CClassFactory, 0);

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CClassFactory_CreateInstance (IClassFactory* This, IUnknown *pUnkOuter, REFIID riid, void **ppvObject)
{
    CClassFactory* cf;

    COM_INIT_PTR_ARG (ppvObject);
    COM_METHOD_PROLOGUE (cf, !cf || !cf->intfClassFactory || !cf->objectClass || !riid);

    return COM_CreateInstance (cf->objectClass, pUnkOuter, riid, ppvObject);
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP CClassFactory_LockServer (IClassFactory* This, BOOL fLock)
{
    COM_LockServer ((BOOL)fLock);

    return S_OK;
}

/* -------------------------------------------------------------------------- */

static HRESULT COM_CTOR_NAME (CClassFactory) (COM_OBJECT* object)
{
    return S_OK;
}

/* -------------------------------------------------------------------------- */

static void COM_DTOR_NAME (CClassFactory) (COM_OBJECT* object)
{
    CClassFactory* cf;
    COM_RUNTIME_CLASS* runtimeClass;

    cf = (CClassFactory*)object;

    COM_EnterObject();

    if (cf && cf->objectClass)
    {
        runtimeClass = (COM_RUNTIME_CLASS*)cf->objectClass->runtimeClass;
        if (runtimeClass)
            runtimeClass->classFactory = NULL;
    }

    COM_LeaveObject();
}

/* -------------------------------------------------------------------------- */
/*                                                 */
/* -------------------------------------------------------------------------- */

BOOL COM_AddItem (COM_LIST* list, COM_LIST_ITEM* newItem)
{
    if (!list || !newItem)
        return FALSE;

    newItem->next = NULL;

    newItem->prev = list->tail;
    list->tail = newItem;

    if (newItem->prev)
        newItem->prev->next = newItem;
    else
        list->head = newItem;

    return TRUE;
}

/* -------------------------------------------------------------------------- */

BOOL COM_InsertItemBefore (COM_LIST* list, COM_LIST_ITEM* item, COM_LIST_ITEM* newItem)
{
    if (!list || !item || !newItem)
        return FALSE;

    newItem->next = item;
    newItem->prev = item->prev;
    item->prev = newItem;

    if (!item->prev)
        list->head = item;

    return TRUE;
}

/* -------------------------------------------------------------------------- */

BOOL COM_InsertItemAfter (COM_LIST* list, COM_LIST_ITEM* item, COM_LIST_ITEM* newItem)
{
    if (!list || !item || !newItem)
        return FALSE;

    newItem->prev = item;
    newItem->next = item->next;
    item->next = newItem;

    if (!item->next)
        list->tail = item;

    return TRUE;
}

/* -------------------------------------------------------------------------- */

void COM_RemoveItem (COM_LIST* list, COM_LIST_ITEM* item)
{
    if (list && item)
    {
        COM_InvalidateListIterators (list, item);

        if (item->next)
            item->next->prev = item->prev;
        else
            list->tail = item->prev;

        if (item->prev)
            item->prev->next = item->next;
        else
            list->head = item->next;

        item->prev = NULL;
        item->next = NULL;

        if (list->Free)
            list->Free (list, item);
    }
}

/* -------------------------------------------------------------------------- */

/*
    COM_ClearList

          list.
*/
void COM_ClearList (COM_LIST* list)
{
    COM_LIST_ITEM *item, *next;

    if (list)
    {
        if (list->iterators)
        {
            COM_InvalidateListIterators (list, NULL);
            COM_ClearList (list->iterators);
            _Delete (list->iterators, COM_LIST);
            list->iterators = NULL;
        }

        item = list->head;

        list->head = NULL;
        list->tail = NULL;

        if (list->Free)
        {
            while (item)
            {
                next = item->next;
                
                item->prev = NULL;
                item->next = NULL;

                list->Free (list, item);

                item = next;
            }
        }
        else
        {
            while (item)
            {
                next = item->next;
                
                item->prev = NULL;
                item->next = NULL;

                item = next;
            }
        }
    }
}

/* -------------------------------------------------------------------------- */

static void FreeListIterator (COM_LIST* list, COM_LIST_ITEM* item)
{
    if (item)
        _Delete (item, COM_LIST_ITER);
}

/* -------------------------------------------------------------------------- */

/*
    COM_IsValidListIterator

     TRUE,   iter   list.
*/
BOOL COM_IsValidListIterator (COM_LIST* list, COM_LIST_ITER* iter)
{
    COM_LIST_ITEM* item;

    if (list && list->iterators && iter)
    {
        item = (COM_LIST_ITEM*)iter;
        while (item->prev)
            item = item->prev;

        return (item == list->iterators->head);
    }

    return FALSE;
}

/* -------------------------------------------------------------------------- */

/*
    COM_CreateListIterator

      ,     list.

        NULL    ( ).
*/
COM_LIST_ITER* COM_CreateListIterator (COM_LIST* list)
{
    COM_LIST_ITER* iter;

    if (!list)
        return NULL;

    if (!list->iterators)
    {
        list->iterators = _New (COM_LIST);
        if (!list->iterators)
            return NULL;

        list->iterators->Free = FreeListIterator;
    }

    iter = _New (COM_LIST_ITER);
    if (!iter)
    {
        if (!list->iterators->head)
        {
            _Delete (list->iterators, COM_LIST);
            list->iterators = NULL;
        }

        return NULL;
    }

    iter->item = COM_LIST_ITER_FIRST_ITEM;

    COM_AddItem (list->iterators, &iter->iterItem);

    return iter;
}

/* -------------------------------------------------------------------------- */

/*
    COM_CopyListIterator

      ,       list,   iter.

        NULL    ( 
       iter).
*/
COM_LIST_ITER* COM_CopyListIterator (COM_LIST* list, COM_LIST_ITER* iter)
{
    COM_LIST_ITER* newIter;

    if (!COM_IsValidListIterator (list, iter))
        return NULL;

    newIter = COM_CreateListIterator (list);

    if (newIter)
        newIter->item = iter->item;

    return newIter;
}

/* -------------------------------------------------------------------------- */

/*
    COM_FreeListIterator

     ,   COM_CreateListIterator()  
    COM_CopyListIterator().
*/
void COM_FreeListIterator (COM_LIST* list, COM_LIST_ITER* iter)
{
    if (COM_IsValidListIterator (list, iter))
    {
        iter->item = NULL;

        COM_RemoveItem (list->iterators, &iter->iterItem);

        if (!list->iterators->head)
        {
            _Delete (list->iterators, COM_LIST);
            list->iterators = NULL;
        }
    }
}

/* -------------------------------------------------------------------------- */

/*
    COM_InvalidateListIterators

      ,    item, .
     item  NULL,    list   
    .

         COM_RemoveItem()  
    COM_ClearList().

      ,     -   
    ( ).
*/
void COM_InvalidateListIterators (COM_LIST* list, COM_LIST_ITEM* item/*=NULL*/)
{
    COM_LIST_ITER* iter;

    if (list && list->iterators)
    {
        iter = (COM_LIST_ITER*)list->iterators->head;

        if (item)
        {
            while (iter)
            {
                if (iter->item == item)
                    iter->item = item->next;

                iter = (COM_LIST_ITER*)iter->iterItem.next;
            }
        }
        else
        {
            while (iter)
            {
                iter->item = NULL;

                iter = (COM_LIST_ITER*)iter->iterItem.next;
            }
        }
    }
}

/* -------------------------------------------------------------------------- */
/*                      AVL-                                  */
/* -------------------------------------------------------------------------- */

static void AVL_RebalanceTree (COM_TREE* tree)
{

}

/* -------------------------------------------------------------------------- */

static DWORD AVL_NodeHeight (COM_TREE_NODE* node)
{
    DWORD leftHeight, rightHeight;

    if (!node)
        return 0;

    leftHeight = AVL_NodeHeight (node->left);
    rightHeight = AVL_NodeHeight (node->right);

    return ((rightHeight > leftHeight) ? rightHeight : leftHeight) + 1;
}

/* -------------------------------------------------------------------------- */

static int AVL_NodeBalanceFactor (COM_TREE_NODE* node)
{
    DWORD leftHeight, rightHeight;

    if (!node)
        return 0;

    leftHeight = AVL_NodeHeight (node->left);
    rightHeight= AVL_NodeHeight (node->right);

    return (int)(rightHeight - leftHeight);
}

/* -------------------------------------------------------------------------- */

/*
        [1]                 [3]
     [2]   [3]     =>    [1]   [5]
         [4] [5]       [2] [4]
*/
static BOOL AVL_SingleRotateRight (COM_TREE* tree, COM_TREE_NODE* node)
{
    COM_TREE_NODE* pivot;
    BOOL isLeft;

    if (!node->right)
    {
        ErrMsg ("AVL: rotate right failed: node->right is NULL");

        return FALSE;
    }

    isLeft = (node->parent && node == node->parent->left);

    pivot = node->right;
    node->right = pivot->left;
    pivot->left = node;
    pivot->parent = node->parent;
    node->parent = pivot;

    if (node->right)
        node->right->parent = node;

    if (pivot->parent)
    {
        if (isLeft)
            pivot->parent->left = pivot;
        else
            pivot->parent->right = pivot;
    }
    else
        tree->root = pivot;
    
    return TRUE;
}

/* -------------------------------------------------------------------------- */

static BOOL AVL_SingleRotateLeft (COM_TREE* tree, COM_TREE_NODE* node)
{
    COM_TREE_NODE* pivot;
    BOOL isLeft;

    if (!node->left)
    {
        ErrMsg ("AVL: rotate left failed: node->left is NULL");

        return FALSE;
    }

    isLeft = (node->parent && node == node->parent->left);

    pivot = node->left;
    node->left = pivot->right;
    pivot->right = node;
    pivot->parent = node->parent;
    node->parent = pivot;

    if (node->left)
        node->left->parent = node;

    if (pivot->parent)
    {
        if (isLeft)
            pivot->parent->left = pivot;
        else
            pivot->parent->right = pivot;
    }
    else
        tree->root = pivot;

    return TRUE;
}

/* -------------------------------------------------------------------------- */

static BOOL AVL_DoubleRotateRight (COM_TREE* tree, COM_TREE_NODE* node)
{
    if (!node->right || !node->right->left)
    {
        ErrMsg ("AVL: double rotate right failed: node->%s is NULL", (node->right ? "right->Left" : "right"));

        return FALSE;
    }

    if (!AVL_SingleRotateLeft (tree, node->right))
        return FALSE;
        
    return AVL_SingleRotateRight (tree, node);
}

/* -------------------------------------------------------------------------- */

static BOOL AVL_DoubleRotateLeft (COM_TREE* tree, COM_TREE_NODE* node)
{
    if (!node->left || !node->left->right)
    {
        ErrMsg ("AVL: double rotate left failed: node->%s is NULL", (node->left ? "left->right" : "left"));

        return FALSE;
    }

    if (!AVL_SingleRotateRight (tree, node->left))
        return FALSE;
        
    return AVL_SingleRotateLeft (tree, node);
}

/* -------------------------------------------------------------------------- */

/*
    AVL_RebalanceNode

       node.

      ,       
     FALSE.
*/
static BOOL AVL_RebalanceNode (COM_TREE* tree, COM_TREE_NODE* node)
{
    int bf, errorCase;
    
    errorCase= 0;

    bf = node->bf;

    if (bf == 2)
    {
        if (node->right)
        {
            if (node->right->bf == 0)
            {
                AVL_SingleRotateRight (tree, node);

                node->bf = 1;
                node->parent->bf = -1; // node->right  
            }
            else if (node->right->bf == 1)
            {
                AVL_SingleRotateRight (tree, node);

                node->bf = 0;
                node->parent->bf = 0; // node->right  
            }
            else if (node->right->bf == -1)
            {
                if (AVL_DoubleRotateRight (tree, node))
                {
                    bf = node->parent->bf; /* node->right->left   */
                    if (bf == -1)
                    {
                        node->bf = 0;                /* node   */
                        node->parent->bf = 0;         /* node->right->left   */
                        node->parent->right->bf = 1;   /* node->right   */
                    }
                    else if (bf == 1)
                    {
                        node->bf = -1;               /* node   */
                        node->parent->bf = 0;         /* node->right->left   */
                        node->parent->right->bf = 0;   /* node->right   */
                    }
                    else if (bf == 0)
                    {
                        node->bf = 0;                /* node   */
                        node->parent->bf = 0;         /* node->right->left   */
                        node->parent->right->bf = 0;   /* node->right   */
                    }
                    else
                        errorCase = 1;
                }
                else
                    errorCase = 2;
            }
            else
                errorCase = 3;
        }
        else
            errorCase = 4;
    }
    else if (bf == -2)
    {
        if (node->left)
        {
            if (node->left->bf == 0)
            {
                AVL_SingleRotateLeft (tree, node);

                node->bf = -1;
                node->parent->bf = 1;   /* node->left   */
            }
            else if (node->left->bf == -1)
            {
                AVL_SingleRotateLeft (tree, node);

                node->bf = 0;
                node->parent->bf = 0;   /* node->left   */
            }
            else if (node->left->bf == 1)
            {
                if (AVL_DoubleRotateLeft (tree, node))
                {
                    bf = node->parent->bf; /* node->left->right   */
                    if (bf == 1)
                    {
                        node->bf = 0;                   /* node   */
                        node->parent->bf = 0;           /* node->left->right   */
                        node->parent->left->bf = -1;    /* node->left   */
                    }
                    else if (bf == -1)
                    {
                        node->bf = 1;                   /* node   */
                        node->parent->bf = 0;           /* node->left->right   */
                        node->parent->left->bf = 0;     /* node->left   */
                    }
                    else if (bf == 0)
                    {
                        node->bf = 0;                   /* node   */
                        node->parent->bf = 0;           /* node->left->right   */
                        node->parent->left->bf = 0;     /* node->left   */
                    }
                    else
                        errorCase = 5;
                }
                else
                    errorCase = 6;
            }
            else
                errorCase = 7;
        }
        else
            errorCase = 8;
    }
    else if (bf > -2 && bf < 2)
    {
        /*    */
    }
    else
        errorCase = 9;

    if (errorCase)
    {
        ErrMsg ("AVL: tree is corrupted (%d). Rebuilding tree...", errorCase);

        return FALSE;
    }

    return TRUE;
}

/* -------------------------------------------------------------------------- */

static void AVL_OnNodeAdded (COM_TREE* tree, COM_TREE_NODE* node, BOOL isLeft)
{
    COM_TREE_NODE* parent;
    int bf;

    while (node)
    {
        parent = node->parent;
        bf = (isLeft ? --node->bf : ++node->bf);

        if (bf == 0)
            break;

        if (bf < -2 || bf > 2)
        {
            ErrMsg ("AVL: tree is corrupted (node balance factor is %d). Rebuilding tree...", bf);
            AVL_RebalanceTree (tree);
            return;
        }

        if (bf < -1 || bf > 1)
        {
            AVL_RebalanceNode (tree, node);
            break;
        }

        isLeft = (parent && (node == parent->left));
        node = parent;
    }
}

/* -------------------------------------------------------------------------- */

static BOOL AVL_ValidateNode (COM_TREE_NODE* node)
{
    if (node->left && !AVL_ValidateNode (node->left))
        return FALSE;

    if (node->right && !AVL_ValidateNode (node->right))
        return FALSE;

    return (node->bf == AVL_NodeBalanceFactor (node));
}

/* -------------------------------------------------------------------------- */

static void AVL_OnNodeRemoved (COM_TREE* tree, COM_TREE_NODE* node, BOOL isLeft)
{
    COM_TREE_NODE* parent;
    int bf, delta;

    if (node)
        delta = (isLeft ? 1 : -1);

    while (node)
    {
        parent = node->parent;
        isLeft = (parent && node == parent->left);
        bf = (node->bf += delta);

        if (bf == -1 || bf == 1)
            break;

        if (bf < -2 || bf > 2)
        {
            ErrMsg ("AVL: tree is corrupted (node balance factor is %d). Rebuilding tree...", bf);
            AVL_RebalanceTree (tree);
            return;
        }

        if (bf == -2 || bf == 2)
        {
            if (!AVL_RebalanceNode (tree, node))
                return;

            if (parent)
            {
                node = (isLeft ? parent->left : parent->right);
                if (node->bf != 0)
                {
                    /*   */
                    break;
                }
            }
        }

        delta = (isLeft ? 1 : -1);

        node = parent;
    }
}

/* -------------------------------------------------------------------------- */

static void AVL_SwapNodes (COM_TREE* tree, COM_TREE_NODE* node1, COM_TREE_NODE* node2)
{
    COM_TREE_NODE* tmp;

    if (node1 == node2)
        return;

    /*    , node1    . */
    
    if (node1 == node2->left || node1 == node2->right)
    {
        tmp = node1;
        node1 = node2;
        node2 = tmp;
    }

    /*    */

    if (node1->parent)
    {
        if (node1 == node1->parent->left)
            node1->parent->left = node2;
        else
            node1->parent->right = node2;
    }
    else
        tree->root = node2;

    if (node2 == node1->left || node2 == node1->right)
    {
        node2->parent = node1->parent;
        node1->parent = node2;
    }
    else
    {
        if (node2->parent)
        {
            if (node2 == node2->parent->left)
                node2->parent->left = node1;
            else
                node2->parent->right = node1;
        }
        else
            tree->root = node1;

        tmp = node1->parent;
        node1->parent = node2->parent;
        node2->parent = tmp;
    }

    /*    */

    if (node1->left == node2)
    {
        if (node2->left)
            node2->left->parent = node1;

        node1->left = node2->left;
        node2->left = node1;
    }
    else
    {
        if (node1->left)
            node1->left->parent = node2;

        if (node2->left)
            node2->left->parent = node1;

        tmp = node1->left;
        node1->left = node2->left;
        node2->left = tmp;
    }

    /*    */

    if (node1->right == node2)
    {
        if (node2->right)
            node2->right->parent = node1;

        node1->right = node2->right;
        node2->right = node1;
    }
    else
    {
        if (node1->right)
            node1->right->parent = node2;

        if (node2->right)
            node2->right->parent = node1;

        tmp = node1->right;
        node1->right = node2->right;
        node2->right = tmp;
    }

    node1->bf ^= node2->bf;
    node2->bf ^= node1->bf;
    node1->bf ^= node2->bf;
}

/* -------------------------------------------------------------------------- */

static void AVL_RemoveNode (COM_TREE* tree, COM_TREE_NODE* node)
{
    COM_TREE_NODE *parent, *child;
    BOOL isLeft;

    if (!node)
        return;

    if (node->left && node->right)
    {
        child = node->right;
        while (child->left)
            child = child->left;

        AVL_SwapNodes (tree, node, child);
        AVL_RemoveNode (tree, node);

        return;
    }

    parent = node->parent;
    isLeft = (parent && node == parent->left);

    child = (node->left) ? node->left : node->right;

    if (child)
        child->parent = parent;

    if (parent)
    {
        if (isLeft)
            parent->left = child;
        else
            parent->right = child;

        node->parent = NULL;
    }
    else
        tree->root = child;

    node->left = NULL;
    node->right = NULL;

    if (tree->Free)
        tree->Free (tree, node);

    if (parent)
        AVL_OnNodeRemoved (tree, parent, isLeft);
}

/* -------------------------------------------------------------------------- */

static void AVL_RemoveTree (COM_TREE* tree, COM_TREE_NODE* node)
{
    if (node->left)
    {
        AVL_RemoveTree (tree, node->left);
        node->left = NULL;
    }

    if (node->right)
    {
        AVL_RemoveTree (tree, node->right);
        node->right = NULL;
    }

    if (node->parent)
    {
        if (node == node->parent->left)
            node->parent->left = NULL;
        else if (node == node->parent->right)
            node->parent->right = NULL;

        node->parent = NULL;
    }
    else if (node == tree->root)
        tree->root = NULL;

    if (tree->Free)
        tree->Free (tree, node);
}

/* -------------------------------------------------------------------------- */

/*
    COM_FirstNode

       .
*/
COM_TREE_NODE* COM_FirstNode (COM_TREE* tree)
{
    COM_TREE_NODE* node;

    if (!tree || !tree->root)
        return NULL;

    node = tree->root;

    while (node && node->left)
        node = node->left;

    return node;
}

/* -------------------------------------------------------------------------- */

/*
    COM_LastNode

       .
*/
COM_TREE_NODE* COM_LastNode (COM_TREE* tree)
{
    COM_TREE_NODE* node;

    if (!tree || !tree->root)
        return NULL;

    node = tree->root;

    while (node && node->right)
        node = node->right;

    return node;
}

/* -------------------------------------------------------------------------- */

/*
    COM_PrevNode

           .
*/
COM_TREE_NODE* COM_PrevNode (COM_TREE* tree, COM_TREE_NODE* node)
{
    COM_TREE_NODE* prev;

    if (!tree || !node)
        return NULL;

    prev = node->left;

    if (prev == NULL)
    {
        if (node->parent && node == node->parent->left)
        {
            prev = node->parent;
            while (prev)
            {
                if (!prev->parent || prev != prev->parent->left)
                {
                    prev = prev->parent;
                    break;
                }

                prev = prev->parent;
            }
        }
        else
            prev = node->parent;
    }
    else
    {
        while (prev->right)
            prev = prev->right;
    }

    return prev;
}

/* -------------------------------------------------------------------------- */

/*
    COM_NextNode

           .
*/
COM_TREE_NODE* COM_NextNode (COM_TREE* tree, COM_TREE_NODE* node)
{
    COM_TREE_NODE* next;

    if (!tree || !node)
        return NULL;

    next = node->right;

    if (next == NULL)
    {
        if (!node->parent || node != node->parent->left)
        {
            next = node->parent;
            while (next)
            {
                if (next->parent && next == next->parent->left)
                {
                    next = next->parent;
                    break;
                }

                next = next->parent;
            }
        }
        else
            next = node->parent;
    }
    else
    {
        while (next->left)
            next = next->left;
    }

    return next;
}

/* -------------------------------------------------------------------------- */

/*
    COM_AddNode

       .

     FALSE,   key  .
*/
BOOL COM_AddNode (COM_TREE* tree, LPCVOID key, COM_TREE_NODE* node)
{
    COM_TREE_NODE* parent;
    int result;

    if (!tree || !tree->Compare || !node)
        return FALSE;

    node->left = NULL;
    node->right = NULL;
    node->bf = 0;

    if (!tree->root)
    {
        node->parent = NULL;
        tree->root = node;

        return TRUE;
    }

    parent = tree->root;

    while (TRUE)
    {
        result = tree->Compare (tree, key, parent);

        if (result < 0)
        {
            if (!parent->left)
            {
                parent->left = node;
                node->parent = parent;
                break;
            }

            parent = parent->left;
        }
        else if (result > 0)
        {
            if (!parent->right)
            {
                parent->right = node;
                node->parent = parent;
                break;
            }

            parent = parent->right;
        }
        else
        {
            /*    */
            return FALSE;
        }
    }

    AVL_OnNodeAdded (tree, parent, node == node->parent->left);

    return TRUE;
}

/* -------------------------------------------------------------------------- */

/*
    COM_RemoveKey

     ,  key.
*/
void COM_RemoveKey (COM_TREE* tree, LPCVOID key)
{
    COM_TREE_NODE* node;

    if (!tree || !tree->root)
        return;

    node = COM_FindNode (tree, key, NULL);

    if (node)
    {
        COM_InvalidateTreeIterators (tree, node);
        AVL_RemoveNode (tree, node);
    }
}

/* -------------------------------------------------------------------------- */

/*
    COM_FindNode

      ,  key.

     ,  key ,  parent    
     ,  -    key .

     parent   NULL.
*/
COM_TREE_NODE* COM_FindNode (COM_TREE* tree, LPCVOID key, COM_TREE_NODE** parent/*=NULL*/)
{
    COM_TREE_NODE *node, *_parent;
    int result;

    if (parent)
        *parent = NULL;

    if (!tree || !tree->Compare || !tree->root)
        return NULL;

    _parent = NULL;
    node = tree->root;

    while (node)
    {
        result = tree->Compare (tree, key, node);

        if (result == 0)
        {
            _parent = node->parent;
            break;
        }
        else
        {
            _parent = node;

            if (result < 0)
                node = node->left;
            else if (result > 0)
                node = node->right;
        }
    }

    if (parent)
        *parent = _parent;

    return node;
}

/* -------------------------------------------------------------------------- */

/*
    COM_ClearTree

          tree.
*/
void COM_ClearTree (COM_TREE* tree)
{
    if (tree)
    {
        if (tree->iterators)
        {
            COM_InvalidateTreeIterators (tree, NULL);
            COM_ClearList (tree->iterators);
            _Delete (tree->iterators, sizeof(COM_LIST));
            tree->iterators = NULL;
        }

        if (tree->root)
        {
            AVL_RemoveTree (tree, tree->root);
            tree->root = NULL;
        }
    }
}

/* -------------------------------------------------------------------------- */

static void FreeTreeIterator (COM_LIST* list, COM_LIST_ITEM* item)
{
    if (item)
        _Delete (item, COM_TREE_ITER);
}

/* -------------------------------------------------------------------------- */

/*
    COM_IsValidTreeIterator

     TRUE,   iter   tree.
*/
BOOL COM_IsValidTreeIterator (COM_TREE* tree, COM_TREE_ITER* iter)
{
    COM_LIST_ITEM* item;

    if (tree && tree->iterators && iter)
    {
        item = (COM_LIST_ITEM*)&iter->item;
        while (item->prev)
            item = item->prev;

        return (item == tree->iterators->head);
    }

    return FALSE;
}

/* -------------------------------------------------------------------------- */

/*
    COM_CreateTreeIterator

      ,      tree.

        NULL    ( ).
*/
COM_TREE_ITER* COM_CreateTreeIterator (COM_TREE* tree)
{
    COM_TREE_ITER* iter;

    if (!tree)
        return NULL;

    if (!tree->iterators)
    {
        tree->iterators = _New (COM_LIST);
        if (!tree->iterators)
            return NULL;

        tree->iterators->Free = FreeTreeIterator;
    }

    iter = _New (COM_TREE_ITER);
    if (!iter)
    {
        if (!tree->iterators->head)
        {
            _Delete (tree->iterators, COM_LIST);
            tree->iterators = NULL;
        }

        return NULL;
    }

    iter->node = COM_TREE_ITER_FIRST_NODE;

    COM_AddItem (tree->iterators, &iter->item);

    return iter;
}

/* -------------------------------------------------------------------------- */

/*
    COM_CopyTreeIterator

      ,       tree,   iter.

        NULL    ( 
       iter).
*/
COM_TREE_ITER* COM_CopyTreeIterator (COM_TREE* tree, COM_TREE_ITER* iter)
{
    COM_TREE_ITER* newIter;

    if (!COM_IsValidTreeIterator (tree, iter))
        return NULL;

    newIter = COM_CreateTreeIterator (tree);

    if (newIter)
        newIter->node = iter->node;

    return newIter;
}

/* -------------------------------------------------------------------------- */

/*
    COM_FreeTreeIterator

     ,   COM_CreateTreeIterator()  
    COM_CopyTreeIterator().
*/
void COM_FreeTreeIterator (COM_TREE* tree, COM_TREE_ITER* iter)
{
    if (COM_IsValidTreeIterator (tree, iter))
    {
        iter->node = NULL;

        COM_RemoveItem (tree->iterators, &iter->item);

        if (!tree->iterators->head)
        {
            _Delete (tree->iterators, COM_LIST);
            tree->iterators = NULL;
        }
    }
}

/* -------------------------------------------------------------------------- */

/*
    COM_InvalidateTreeIterators

      ,    node, .
     node  NULL,    tree   
    .

         COM_RemoveKey()  
    COM_ClearTree().

      ,     -   
    ( ).
*/
void COM_InvalidateTreeIterators (COM_TREE* tree, COM_TREE_NODE* node/*=NULL*/)
{
    COM_TREE_ITER* iter;

    if (tree && tree->iterators)
    {
        iter = (COM_TREE_ITER*)tree->iterators->head;

        if (node)
        {
            while (iter)
            {
                if (iter->node == node)
                    iter->node = COM_NextNode (tree, iter->node);

                iter = (COM_TREE_ITER*)iter->item.next;
            }
        }
        else
        {
            while (iter)
            {
                iter->node = NULL;

                iter = (COM_TREE_ITER*)iter->item.next;
            }
        }
    }
}

/* -------------------------------------------------------------------------- */
/*                                                   */
/* -------------------------------------------------------------------------- */

/*
    COM_GetEvents

     ,     .
*/
COM_EVENTS* COM_GetEvents (void)
{
    return &g_Server.events;
}

/* -------------------------------------------------------------------------- */

/*
    COM_SetEventsParam

      ,    
    .
*/
void COM_SetEventsParam (LPVOID param)
{
    g_Server.eventsParam = param;
}

/* -------------------------------------------------------------------------- */

/*
    COM_GetStats

      stats  .
*/
void COM_GetStats (COM_STATS* stats)
{
    FILETIME now, start;

    if (stats)
    {
        _Copy (&stats, &g_Server.stats, sizeof(COM_STATS));

        if (stats->isActive)
        {
            GetSystemTimeAsFileTime (&now);
            SystemTimeToFileTime (&g_Server.stats.startTime, &start);
            
            now.dwHighDateTime -= start.dwHighDateTime;
            if (now.dwLowDateTime < start.dwLowDateTime)
                now.dwHighDateTime--;
            now.dwLowDateTime -= start.dwLowDateTime;

            FileTimeToSystemTime (&now, &stats->runTime);
        }
    }
}

/* -------------------------------------------------------------------------- */

/*
    COM_EnumRunningClass

      callback   ,     
    .
*/
void COM_EnumRunningClasses (int (*callback)(const COM_CLASS*, LPVOID), LPVOID param)
{
    COM_RUNTIME_CLASS* item;

    if (callback)
    {
        COM_EnterObject();

        item = (COM_RUNTIME_CLASS*)g_Server.runningClassList.head;
        while (item)
        {
            if (item->_class)
                if (callback (item->_class, param) != COM_ENUM_CONTINUE)
                    break;

            item = (COM_RUNTIME_CLASS*)item->listItem.next;
        }

        COM_LeaveObject();
    }
}

/* -------------------------------------------------------------------------- */

/*
    COM_EnumRunningObjects

      callback    .
*/
void COM_EnumRunningObjects (int (*callback)(COM_OBJECT*, LPVOID), LPVOID param)
{
    COM_RUNTIME_CLASS* classItem;
    COM_LIST_ITEM* objectItem;
    COM_OBJECT* object;
    int result;

    if (callback)
    {
        COM_EnterObject();

        classItem = (COM_RUNTIME_CLASS*)g_Server.runningClassList.head;
        while (classItem)
        {
            objectItem = classItem->objectList.head;
            while (objectItem)
            {
                object = OBJECT_FROM_FIELD (objectItem, COM_OBJECT, listItem);
                result = callback (object, param);

                if (result == COM_ENUM_SKIP_CLASS)
                    break;

                if (result != COM_ENUM_CONTINUE)
                {
                    COM_LeaveObject();

                    return;
                }

                objectItem = objectItem->next;
            }

            classItem = (COM_RUNTIME_CLASS*)classItem->listItem.next;
        }

        COM_LeaveObject();
    }
}

/* -------------------------------------------------------------------------- */
/*                          GUID                             */
/* -------------------------------------------------------------------------- */

/*
    COM_IsEqualGUID

       GUID.

     TRUE,    .
*/
BOOL COM_IsEqualGUID (LPCGUID guid1, LPCGUID guid2)
{
    return (COM_CompareGUID (guid1, guid2) == 0);
}

/* -------------------------------------------------------------------------- */

/*
    COM_CompareGUID

       GUID.

    :
        -1  - guid1 < guid2
        1   - guid1 > guid2
        0   - guid1 == guid2
*/
int COM_CompareGUID (LPCGUID guid1, LPCGUID guid2)
{
    if (guid1 == guid2)
        return 0;

    if (guid1 && guid2)
        return _Compare (guid1, guid2, sizeof(GUID));

    if (!guid1 && guid2)
        return -1;

    /* guid1 && !guid2 */

    return 1;
}

/* -------------------------------------------------------------------------- */

/*
    COM_CopyGUID

       src  dst.
    
     src  NULL,  dst  .
*/
void COM_CopyGUID (LPGUID dest, LPCGUID src/*=NULL*/)
{
    if (dest)
    {
        if (src)
            _Copy (dest, src, sizeof(GUID));
        else
            _Clear (dest, sizeof(GUID));
    }
}

/* -------------------------------------------------------------------------- */

static STDMETHODIMP GUID_QueryInterface (IGuid* This, REFIID iid, LPVOID* ppvObject)
{
    return E_NOTIMPL;
}

/* -------------------------------------------------------------------------- */

static ULONG WINAPI GUID_AddRef (IGuid* This)
{
    IGuidObject* object;

    object = (IGuidObject*)This;
    if (!object)
        return 0;

    return IncrementCounter ("IGuid", "refCount", &object->refCount);
}

/* -------------------------------------------------------------------------- */

static ULONG WINAPI GUID_Release (IGuid* This)
{
    IGuidObject* object;
    DWORD refCount;

    object = (IGuidObject*)This;

    if (!object)
        return 0;

    refCount = DecrementCounter (&object->refCount);
    if (refCount == 0)
    {
        COM_EnterGuid();

        if (object->node)
        {
            object->node = NULL;

            COM_RemoveKey (&g_Server.guidTree, &object->guid);
        }

        COM_LeaveGuid();

        _Delete (object, IGuidObject);
    }

    return refCount;
}

/* -------------------------------------------------------------------------- */

static REFGUID WINAPI GUID_Value (IGuid* This)
{
    IGuidObject* object;

    object = (IGuidObject*)This;

    return (object ? &object->guid : NULL);
}

/* -------------------------------------------------------------------------- */

static int WINAPI GUID_Compare (IGuid* This, REFGUID guid)
{
    IGuidObject* object;

    object = (IGuidObject*)This;

    return ((object && guid) ? COM_CompareGUID (&object->guid, guid) : -1);
}

/* -------------------------------------------------------------------------- */

static BOOL WINAPI GUID_Equal (IGuid* This, REFGUID guid)
{
    IGuidObject* object;

    object = (IGuidObject*)This;

    return ((object && guid) ? COM_IsEqualGUID (&object->guid, guid) : FALSE);
}

/* -------------------------------------------------------------------------- */

COM_DECLARE_VTABLE (GUID, IGuid)
{
    GUID_QueryInterface,
    GUID_AddRef,
    GUID_Release,
    GUID_Value,
    GUID_Compare,
    GUID_Equal,
};

/* -------------------------------------------------------------------------- */

IGuid* COM_GetGuidInterface (LPCGUID guid)
{
    IGuidObject* guidObject;
    GUID_NODE* node;

    guidObject = NULL;

    if (guid)
    {
        COM_EnterGuid();

        node = (GUID_NODE*)COM_FindNode (&g_Server.guidTree, guid, NULL);
        if (node)
        {
            guidObject = node->object;
            IGuid_AddRef ((IGuid*)guidObject);
        }
        else
        {
            guidObject = _New (IGuidObject);
            if (guidObject)
            {
                guidObject->object.lpVtbl = & COM_VTABLE_NAME (GUID, IGuid);
                guidObject->refCount = 1;

                COM_CopyGUID (&guidObject->guid, guid);

                node = _New (GUID_NODE);
                if (node)
                {
                    node->object = guidObject;

                    if (!COM_AddNode (&g_Server.guidTree, guid, (COM_TREE_NODE*)node))
                    {
                        _Delete (node, GUID_NODE);
                        node = NULL;
                    }
                }

                if (node)
                {
                    guidObject->node = node;
                }
                else
                {
                    _Delete (guidObject, IGuidObject);
                    guidObject = NULL;
                }
            }
        }

        COM_LeaveGuid();
    }

    return (IGuid*)guidObject;
}

/* -------------------------------------------------------------------------- */

static void FreeGuidNode (COM_TREE* tree, COM_TREE_NODE* node)
{
    GUID_NODE* guidNode;
    IGuidObject* guidObject;

    guidNode = (GUID_NODE*)node;
    if (guidNode)
    {
        guidObject = guidNode->object;
        guidNode->object = NULL;

        if (guidObject)
        {
            guidObject->node = NULL;
            IGuid_Release ((IGuid*)guidObject);
        }

        _Delete (guidNode, GUID_NODE);
    }
}

/* -------------------------------------------------------------------------- */

static int CompareGuidNodes (COM_TREE* tree, LPCVOID key, COM_TREE_NODE* node)
{
    LPCGUID guid;
    GUID_NODE* guidNode;

    guid = (LPCGUID)key;
    guidNode = (GUID_NODE*)node;
    if (guid && guidNode && guidNode->object)
        return COM_CompareGUID (guid, &guidNode->object->guid);

    return -1;
}

/* -------------------------------------------------------------------------- */
/*                                                          */
/* -------------------------------------------------------------------------- */

static BOOL IsInitialized (LPCSTR source)
{
    if (!g_Server.stats.isActive)
    {
        if (source)
            ErrMsg ("%s: COM server is not initialized", source);

        return FALSE;
    }

    return TRUE;
}

/* -------------------------------------------------------------------------- */

/*
    COM_Initialize

     .

         .    ,
       COM_Cleanup()   .
*/
BOOL COM_Initialize (void)
{
    int i;

    if (!g_Server.stats.isActive)
    {
        _Clear (&g_Server.stats, sizeof(COM_STATS));

        g_Server.defaultTypeLib = NULL;

        g_Server.typeLibTree.Free = FreeTypeLibNode;
        g_Server.typeLibTree.Compare = CompareTypeLibNodes;

        g_Server.runningClassList.head = NULL;
        g_Server.runningClassList.tail = NULL;
        g_Server.runningClassList.Free = NULL;

        for (i=0; i<NUMBER_OF_SERVER_LOCK_IDS; i++)
            InitializeCriticalSection (&g_Server.locks[i]);

        g_Server.guidTree.Free = FreeGuidNode;
        g_Server.guidTree.Compare = CompareGuidNodes;

        g_Server.stats.threadId = GetCurrentThreadId ();

        GetSystemTime (&g_Server.stats.startTime);

        g_Server.GIT = NULL;

        g_Server.stats.isActive = TRUE;
    }

    return g_Server.stats.isActive;
}

/* -------------------------------------------------------------------------- */

static void CheckRunningObjects (void)
{
    COM_EnterObject();

    if (g_Server.stats.objectCount != 0 || g_Server.runningClassList.head != NULL)
    {
        FIRE_OnMemoryLeak ();            
    }

    COM_LeaveObject();
}

/* -------------------------------------------------------------------------- */

static void CheckRunningGuids (void)
{
    GUID_NODE* guidNode;
    char guidString[256];

    COM_EnterGuid();

    if (g_Server.guidTree.root)
    {
        ErrMsg ("Running GUID objects detected:");

        guidNode = (GUID_NODE*)COM_FirstNode (&g_Server.guidTree);
        while (guidNode)
        {
            if (guidNode->object)
            {
                _StringFromGUID (&guidNode->object->guid, guidString);
                ErrMsg ("{%s} (%u references)", guidString, guidNode->object->refCount);
            }

            guidNode = (GUID_NODE*)COM_NextNode (&g_Server.guidTree, &guidNode->node);
        }
    }

    COM_LeaveGuid();
}

/* -------------------------------------------------------------------------- */

/*
    COM_Cleanup

     ,  .

           , 
        COM_EVENTS::OnMemoryLeak().
*/
void COM_Cleanup (void)
{
    int i;

    if (IsInitialized ("COM_Cleanup()"))
    {
        CheckRunningGuids();
        CheckRunningObjects();

        if (g_Server.GIT)
        {
            IGlobalInterfaceTable_Release (g_Server.GIT);
            g_Server.GIT = NULL;
        }

        COM_ReleaseTypeLibraries();
        COM_ClearTree (&g_Server.guidTree);

        for (i=0; i<NUMBER_OF_SERVER_LOCK_IDS; i++)
            DeleteCriticalSection (&g_Server.locks[i]);

        g_Server.runningClassList.head = NULL;
        g_Server.runningClassList.tail = NULL;
        g_Server.runningClassList.Free = NULL;

        _Clear (&g_Server.stats, sizeof(COM_STATS));
    }
}

/* -------------------------------------------------------------------------- */

/*
    COM_CanUnloadNow

     TRUE,         0.
*/
BOOL COM_CanUnloadNow (void)
{
    BOOL result;

    result = TRUE;

    if (IsInitialized (NULL))
    {
        COM_EnterObject();

        result = (g_Server.stats.objectCount == 0) && (g_Server.stats.lockCount == 0);

        COM_LeaveObject();

        if (result)
        {
            COM_EnterGuid();
            
            result = g_Server.guidTree.root == NULL;

            COM_LeaveGuid();
        }
    }

    return result;
}

/* -------------------------------------------------------------------------- */

static void COM_EnterLock (int index)
{
    if (IsInitialized (NULL))
        EnterCriticalSection (&g_Server.locks[index]);
}

/* -------------------------------------------------------------------------- */

static void COM_LeaveLock (int index)
{
    if (IsInitialized (NULL))
        LeaveCriticalSection (&g_Server.locks[index]);
}

/* -------------------------------------------------------------------------- */

/*
    COM_EnterEvents
    
       ,      
    COM_EVENTS.
*/
void COM_EnterEvents (void)
{
    COM_EnterLock (EVENTS_LOCK);
}

/* -------------------------------------------------------------------------- */

/*
    COM_LeaveEvents

       ,      
    COM_EVENTS.
*/
void COM_LeaveEvents (void)
{
    COM_LeaveLock (EVENTS_LOCK);
}

/* -------------------------------------------------------------------------- */

/*
    COM_EnterObject
    
       ,       .
*/
void COM_EnterObject (void)
{
    COM_EnterLock (OBJECT_LOCK);
}

/* -------------------------------------------------------------------------- */

/*
    COM_LeaveObject

       ,       .
*/
void COM_LeaveObject (void)
{
    COM_LeaveLock (OBJECT_LOCK);
}

/* -------------------------------------------------------------------------- */

/*
    COM_EnterType
    
       ,       .
*/
void COM_EnterType (void)
{
    COM_EnterLock (TYPE_LOCK);
}

/* -------------------------------------------------------------------------- */

/*
    COM_LeaveType

       ,       .
*/
void COM_LeaveType (void)
{
    COM_EnterLock (TYPE_LOCK);
}

/* -------------------------------------------------------------------------- */

/*
    COM_LockServer

        1   .
*/
void COM_LockServer (BOOL lock)
{
    if (IsInitialized ("COM_LockServer()"))
    {
        COM_EnterObject();

        if (lock)
        {
            IncrementCounter ("COM", "lockCount", &g_Server.stats.lockCount);
        }
        else
        {
            if (DecrementCounter (&g_Server.stats.lockCount) == NEGATIVE_ONE)
            {
                g_Server.stats.lockCount++;
                FIRE_OnNegativeCount ("COM", "lockCount");
            }
        }

        COM_LeaveObject();
    }
}

/* -------------------------------------------------------------------------- */

HRESULT COM_GetGIT (IGlobalInterfaceTable** ppGIT)
{
    HRESULT hr;

    if (!ppGIT)
        return E_INVALIDARG;

    hr = CO_E_NOTINITIALIZED;

    if (IsInitialized ("COM_GetGIT()"))
    {
        COM_EnterObject();

        if (!g_Server.GIT)
            hr = CoCreateInstance (&CLSID_StdGlobalInterfaceTable, NULL, CLSCTX_INPROC_SERVER, &IID_IGlobalInterfaceTable, &g_Server.GIT);

        if (g_Server.GIT)
        {
            IGlobalInterfaceTable_AddRef (g_Server.GIT);
            *ppGIT = g_Server.GIT;
            hr = S_OK;
        }

        COM_LeaveObject();
    }

    return hr;
}

/* -------------------------------------------------------------------------- */
/*                                                 */
/* -------------------------------------------------------------------------- */

BOOL COM_SetDefaultTypeLibrary (ITypeLib* typeLib)
{
    return FALSE;
}

/* -------------------------------------------------------------------------- */

BOOL COM_AddTypeLibrary (ITypeLib* typeLib)
{
    return FALSE;
}

/* -------------------------------------------------------------------------- */

/*
    COM_ReleaseTypeLibraries

       ,  .
*/
void COM_ReleaseTypeLibraries (void)
{
    if (IsInitialized ("COM_ReleaseTypeLibraries()"))
    {
        COM_EnterType();

        if (g_Server.defaultTypeLib)
        {
            FreeTypeLibNode (NULL, &g_Server.defaultTypeLib->node);
            g_Server.defaultTypeLib = NULL;
        }

        COM_ClearTree (&g_Server.typeLibTree);

        COM_LeaveType();
    }
}

/* -------------------------------------------------------------------------- */
/*                                     */
/* -------------------------------------------------------------------------- */

/*
    COM_GetObject

           This,  
     .
*/
COM_OBJECT* _COM_GetObject (LPVOID This)
{
    return (This ? ((IObjectIntf*)This)->object : NULL);
}

/* -------------------------------------------------------------------------- */

/*
    COM_CreateObjectInterface

        .

         .
*/
HRESULT COM_CreateObjectInterface (COM_OBJECT* object, REFIID iid, LPCVOID lpVtbl, LPVOID* ppvObject)
{
    IObjectIntf* intf;
    HRESULT hr;

    COM_INIT_PTR_ARG (ppvObject);

    if (!object || !iid || !lpVtbl)
        return E_INVALIDARG;

    hr = E_OUTOFMEMORY;

    intf = _New (IObjectIntf);
    if (intf)
    {
        intf->lpVtbl = lpVtbl;
        intf->object = object;

        intf->iid = (const IID*)_New (GUID);
        if (intf->iid)
        {
            COM_CopyGUID ((LPGUID)intf->iid, iid);

            *ppvObject = intf;

            hr = S_OK;
        }
    }

    return hr;
}

/* -------------------------------------------------------------------------- */

/*
    COM_DestroyObjectInterface

     ,   COM_CreateObjectInterface().
*/
void COM_DestroyObjectInterface (LPVOID unkn)
{
    IObjectIntf* intf;
    
    intf = (IObjectIntf*)unkn;

    if (intf)
    {
        if (intf->iid)
            _Delete ((LPVOID)intf->iid, GUID);

        if (intf->data)
        {
            COM_ClearTree ((COM_TREE*)intf->data);
            _Delete (intf->data, COM_TREE);
        }

        _Delete (intf, IObjectIntf);
    }
}

/* -------------------------------------------------------------------------- */

HRESULT COM_GetInterfaceTypeInfo (LPVOID unkn, LCID lcid, ITypeInfo** typeInfo)
{
    IObjectIntf* intf;
    INTFTYPEINFONODE tmp, *node;
    HRESULT hr;

    COM_INIT_PTR_ARG (typeInfo);

    intf = (IObjectIntf*)unkn;
    if (!intf || !intf->object)
        return E_INVALIDARG;

    COM_EnterType();

    hr =  E_NOTFOUND;

    if (intf->data)
    {
        tmp.dataNode.type = IDT_TYPEINFO;
        tmp.lcid = lcid;

        node = (INTFTYPEINFONODE*)COM_FindNode ((COM_TREE*)intf->data, &tmp, NULL);
        if (node && node->typeInfo)
        {
            ITypeInfo_AddRef (node->typeInfo);
            *typeInfo = node->typeInfo;
            hr = S_OK;
        }
    }

    COM_LeaveType();

    return hr;
}

/* -------------------------------------------------------------------------- */

/*
    COM_IsAggregatedObject

     TRUE,       .

       InitInstance,    ,
          (InitInstance   
    CLASS_E_NOAGGREGATION).
*/
BOOL COM_IsAggregatedObject (COM_OBJECT* object)
{
    return (object && (object->unknOuter != &object->unknInner));
}

/* -------------------------------------------------------------------------- */

LPCRITICAL_SECTION COM_GetClassLock (COM_OBJECT* object)
{
    const COM_CLASS* _class;
    COM_RUNTIME_CLASS* runtimeClass;

    if (object)
    {
        _class = object->_class;
        while (_class)
        {
            runtimeClass = (COM_RUNTIME_CLASS*)_class->runtimeClass;
            if (runtimeClass && runtimeClass->lock)
                return runtimeClass->lock;

            _class = _class->parent;
        }
    }

    return NULL;
}

/* -------------------------------------------------------------------------- */

BOOL COM_IsClassLock (COM_OBJECT* object, LPCRITICAL_SECTION lock)
{
    const COM_CLASS* _class;
    COM_RUNTIME_CLASS* runtimeClass;

    if (object)
    {
        _class = object->_class;
        while (_class)
        {
            runtimeClass = (COM_RUNTIME_CLASS*)_class->runtimeClass;
            if (runtimeClass && runtimeClass->lock)
                return (lock == runtimeClass->lock);

            _class = _class->parent;
        }
    }

    return FALSE;
}

/* -------------------------------------------------------------------------- */
/*                                                     */
/* -------------------------------------------------------------------------- */

/*
    COM_CreateObject

             
      -.

    :
        _class      -    .
        pUnkOuter   -  IUnknown,    .
        ppvObject   -   ,       
                       (COM_OBJECT*).
*/
HRESULT COM_CreateObject (const COM_CLASS* _class, IUnknown* pUnkOuter, LPVOID* ppvObject)
{
    COM_OBJECT* object;
    COM_RUNTIME_CLASS* runtimeClass;
    DWORD instanceSize;
    HRESULT hr;

    COM_INIT_PTR_ARG (ppvObject);

    if (!IsInitialized ("COM_CreateObject()"))
        return CO_E_NOTINITIALIZED;

    COM_EnterObject();

    if (!IsValidClass (_class))
    {
        COM_LeaveObject();

        return E_INVALIDARG;
    }

    runtimeClass = (COM_RUNTIME_CLASS*)_class->runtimeClass;

    if (runtimeClass)
    {
        if (pUnkOuter && (runtimeClass->flags & COM_FLAGS_NO_AGGREGATION))
        {
            ClassErrMsg (_class, "Class is not aggregatable");

            COM_LeaveObject();

            return CLASS_E_NOAGGREGATION;
        }

        if (!pUnkOuter && (runtimeClass->flags & COM_FLAGS_ONLY_AGGREGATION))
        {
            ClassErrMsg (_class, "Class is only aggregatable");

            COM_LeaveObject();

            return E_FAIL;
        }

        if (runtimeClass->flags & COM_FLAGS_SINGLETON)
        {
            if (pUnkOuter)
            {
                ClassErrMsg (_class, "Class-singleton cannot be aggregated");

                COM_LeaveObject();

                return CLASS_E_NOAGGREGATION;
            }

            if (runtimeClass->singletonInstance)
            {
                IUnknown_AddRef ((IUnknown*)runtimeClass->singletonInstance);
                *ppvObject = runtimeClass->singletonInstance;

                COM_LeaveObject();

                return S_OK;
            }
        }
    }

    instanceSize = CalculateTotalObjectSize (_class);
    if (instanceSize == 0)
    {
        COM_LeaveObject();

        return E_INVALIDARG;
    }

    object = (COM_OBJECT*)_Alloc (instanceSize);
    if (!object)
    {
        COM_LeaveObject();

        return E_OUTOFMEMORY;
    }

    object->unknInner.lpVtbl = &COM_VTABLE_NAME(CUnknown, IUnknown);
    object->unknOuter = pUnkOuter ? pUnkOuter : &object->unknInner;
    object->refCount = 1;
    object->_class = _class;

    hr = _InitObject (object, NULL);
    if (hr != S_OK)
    {
        _Free (object, instanceSize);

        COM_LeaveObject();

        return hr;
    }

    *ppvObject = object;

    if (runtimeClass)
    {
        if (runtimeClass->flags & COM_FLAGS_SINGLETON)
            runtimeClass->singletonInstance = object;

        IncrementCounter (runtimeClass->_class->name, "instanceCount", &runtimeClass->instanceCount);

        if (!runtimeClass->listItem.prev && 
            !runtimeClass->listItem.next && 
            g_Server.runningClassList.head != &runtimeClass->listItem)
        {
            COM_AddItem (&g_Server.runningClassList, &runtimeClass->listItem);
        }

        COM_AddItem (&runtimeClass->objectList, &object->listItem);
    }

    IncrementCounter ("COM", "objectCount", &g_Server.stats.objectCount);

    COM_LeaveObject();

    return S_OK;
}

/* -------------------------------------------------------------------------- */

/*
    COM_CreateInstance

     CoCreateInstance(),      
     CLSID.
*/
HRESULT COM_CreateInstance (const COM_CLASS* _class, IUnknown* pUnkOuter, REFIID riid, LPVOID* ppvObject)
{
    COM_OBJECT* object;
    HRESULT hr;

    COM_INIT_PTR_ARG (ppvObject);

    if (!IsInitialized ("COM_CreateInstance()"))
        return CO_E_NOTINITIALIZED;

    if (!riid)
        return E_INVALIDARG;

    if (pUnkOuter && !COM_IsEqualGUID (riid, &IID_IUnknown))
        return E_NOINTERFACE;

    hr = COM_CreateObject (_class, pUnkOuter, &object);
    if (SUCCEEDED (hr))
    {
        hr = IUnknown_QueryInterface ((IUnknown*)object, riid, ppvObject);
        IUnknown_Release ((IUnknown*)object);
    }

    return hr;
}

/* -------------------------------------------------------------------------- */

/*
    COM_GetClassObject

            .

    :
        _class      -       .
        riid        -    (IID_IUnknown, 
                      IID_IClassFactory).
        ppvObject   -   ,      
                         .
*/
HRESULT COM_GetClassObject (const COM_CLASS* _class, REFIID riid, LPVOID* ppvObject/* IUnknown** */)
{
    CClassFactory* cf;
    IUnknown* intf;
    COM_RUNTIME_CLASS* runtimeClass;
    HRESULT hr;

    COM_INIT_PTR_ARG (ppvObject);

    if (!IsInitialized ("COM_GetClassObject()"))
        return CO_E_NOTINITIALIZED;

    if (!IsValidClass (_class) || !riid)
        return E_INVALIDARG;

    COM_EnterObject();

    runtimeClass = (COM_RUNTIME_CLASS*)_class->runtimeClass;
    if (runtimeClass)
    {
        if (runtimeClass->classFactory)
        {
            hr = IUnknown_QueryInterface (runtimeClass->classFactory, riid, ppvObject);

            COM_LeaveObject();

            return hr;
        }
    }

    hr = COM_CreateInstance (&COM_CLASS_NAME(CClassFactory), NULL, riid, &intf);
    if (FAILED (hr))
    {
        COM_LeaveObject();

        return hr;
    }

    cf = COM_GetObject (CClassFactory, intf);
    if (!cf)
    {
        IUnknown_Release (intf);

        COM_LeaveObject();

        return E_FAIL;
    }

    cf->objectClass = _class;

    if (runtimeClass)
        runtimeClass->classFactory = intf;

    *ppvObject = intf;

    COM_LeaveObject();

    return S_OK;
}

/* -------------------------------------------------------------------------- */

